"""This module implements the default print step."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel, ConfigDict

from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.config.plan import PlanStepConfig
    from ropt.plan import Plan


class DefaultPrintStep(PlanStep):
    """The default print step.

    The print step is used to display messages on the console, using Python's
    `print` function. The message to be printed must be a string evaluated by
    the [`eval`][ropt.plan.Plan.eval] method of the executing
    [`Plan`][ropt.plan.Plan] object. The message should be enclosed in `[[...]]`
    delimiters; if these delimiters are missing, they are implicitly added
    around the message. This format allows for optional interpolation of
    embedded expressions, delimited by `${ ... $}`.

    The print step uses the [`DefaultPrintStepWith`]
    [ropt.plugins.plan._print.DefaultPrintStep.DefaultPrintStepWith]
    configuration class to parse the `with` field of the
    [`PlanStepConfig`][ropt.config.plan.PlanStepConfig] used to specify this step
    in a plan configuration.

    Note: Shorthand configuration
        With this step it is possible to configure the message in shorthand notation.
        Instead of

        ```python
        {
            "run": "print",
            "with": {
                "message": "Hello World!",
            }
        }
        ```

        or

        ```python
        {
            "print": {
                "message": "Hello World!",
            }
        }
        ```

        You can use:

        ```python
        {
            "print":  "Hello World!",
        }
        ```
    """

    class DefaultPrintStepWith(BaseModel):
        """Parameters used by the print step.

        Attributes:
            message: The message to print.
        """

        message: str

        model_config = ConfigDict(
            extra="forbid",
            validate_default=True,
            arbitrary_types_allowed=True,
            frozen=True,
        )

    def __init__(self, config: PlanStepConfig, plan: Plan) -> None:
        """Initialize a default print step.

        Args:
            config: The configuration of the step.
            plan:   The plan that runs this step.
        """
        super().__init__(config, plan)
        if isinstance(config.with_, str):
            self._message = config.with_.strip()
        else:
            _with = self.DefaultPrintStepWith.model_validate(config.with_)
            self._message = _with.message.strip()
        if not (self._message.startswith("[[") and self._message.endswith("]]")):
            self._message = "[[" + self._message + "]]"

    def run(self) -> None:
        """Run the print step."""
        print(self.plan.eval(self._message))  # noqa: T201



"""Definition of the ScriptTask class."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    Optional,
    Tuple,
)

import numpy as np
from numpy.typing import NDArray

from ropt.evaluator.parsl import Task


@dataclass
class ScriptTask(Task):
    """Task class for use by the ScriptBasedOptimizer class."""

    name: str = ""
    job_labels: Tuple[str, ...] = field(default_factory=tuple)
    objective_paths: Tuple[Path, ...] = field(default_factory=tuple)
    constraint_paths: Tuple[Path, ...] = field(default_factory=tuple)
    logged: bool = False
    realization: int = 0
    _objectives: Optional[NDArray[np.float64]] = None
    _constraints: Optional[NDArray[np.float64]] = None

    def _read_file(self, path: Path) -> float:
        try:
            with path.open("r", encoding="utf-8") as file_obj:
                return float(file_obj.read())
        except BaseException as exc:  # noqa: BLE001
            self.exception = exc
            return np.nan

    def _read_results(self) -> None:
        if self.objective_paths:
            self._objectives = np.zeros(len(self.objective_paths), dtype=np.float64)
            for idx, path in enumerate(self.objective_paths):
                self._objectives[idx] = self._read_file(path)
        if self.constraint_paths:
            self._constraints = np.zeros(len(self.constraint_paths), dtype=np.float64)
            for idx, path in enumerate(self.constraint_paths):
                self._constraints[idx] = self._read_file(path)

    def get_objectives(self) -> Optional[NDArray[np.float64]]:
        """Get the objective values.

        Returns:
            The objective values.
        """
        self._read_results()
        return self._objectives

    def get_constraints(self) -> Optional[NDArray[np.float64]]:
        """Get the constraint values.

        Returns:
            The constraint values.
        """
        self._read_results()
        return self._constraints



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

import numpy as np

from ropt import utils
from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import (
    _get_lower_bound_constraint_values,
    _get_lower_bound_violations,
    _get_upper_bound_constraint_values,
    _get_upper_bound_violations,
    _immutable_copy,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig

    from ._function_evaluations import FunctionEvaluations


@dataclass
class BoundConstraints(ResultField):
    """This class stores bound constraint values and violations.

    The following information is stored:

    1. Differences between the current variable and the lower or upper bounds.
    2. Violations of constraints, defined as the absolute value of the
       difference if the constraint is violated, or else zero.
    3. If variable scaling is configured, scaled versions of all values.

    Attributes:
        lower_values:            Lower bound differences.
        lower_violations:        Lower bound violations.
        upper_values:            Upper bound differences.
        upper_violations:        Upper bound violations.
        scaled_lower_values:     Optional scaled lower bound differences.
        scaled_lower_violations: Optional scaled lower bound violations.
        scaled_upper_values:     Optional scaled upper bound differences.
        scaled_upper_violations: Optional scaled upper bound violations.
    """

    lower_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    lower_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    upper_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    upper_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    scaled_lower_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    scaled_lower_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    scaled_upper_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    scaled_upper_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.lower_values = _immutable_copy(self.lower_values)
        self.lower_violations = _immutable_copy(self.lower_violations)
        self.upper_values = _immutable_copy(self.upper_values)
        self.upper_violations = _immutable_copy(self.upper_violations)
        self.scaled_lower_values = _immutable_copy(self.scaled_lower_values)
        self.scaled_lower_violations = _immutable_copy(self.scaled_lower_violations)
        self.scaled_upper_values = _immutable_copy(self.scaled_upper_values)
        self.scaled_upper_violations = _immutable_copy(self.scaled_upper_violations)

    @classmethod
    def create(
        cls, config: EnOptConfig, evaluations: FunctionEvaluations
    ) -> Optional[BoundConstraints]:
        """Add constraint information.

        This factory function creates a `BoundConstraints` object with the
        correct information on bound constraints, based on the given
        evaluations.

        Args:
            config:      The ensemble optimizer configuration object.
            evaluations: An instance of the Evaluations class.

        Returns:
            A newly created BoundConstraints object.
        """

        def _get_unscaled(
            scaled: Optional[NDArray[np.float64]],
        ) -> Optional[NDArray[np.float64]]:
            unscaled = None
            if scaled is not None:
                assert config is not None
                unscaled = utils.scaling.scale_back_variables(
                    config,
                    scaled,
                    correct_offsets=False,
                    axis=-1,
                )
                if unscaled is not None:
                    unscaled.setflags(write=False)
            return unscaled

        kwargs = {}
        if np.any(np.isfinite(config.variables.lower_bounds)):
            lower_values = _get_lower_bound_constraint_values(
                config,
                (
                    evaluations.variables
                    if evaluations.scaled_variables is None
                    else evaluations.scaled_variables
                ),
                axis=-1,
            )
            lower_violations = _get_lower_bound_violations(lower_values)
            unscaled_lower_values = _get_unscaled(lower_values)
            unscaled_lower_violations = _get_unscaled(lower_violations)
            kwargs.update(
                {
                    "lower_values": (
                        lower_values
                        if unscaled_lower_values is None
                        else unscaled_lower_values
                    ),
                    "lower_violations": (
                        lower_violations
                        if unscaled_lower_violations is None
                        else unscaled_lower_violations
                    ),
                    "scaled_lower_values": (
                        None if unscaled_lower_values is None else lower_values
                    ),
                    "scaled_lower_violations": (
                        None if unscaled_lower_violations is None else lower_violations
                    ),
                },
            )
        if np.any(np.isfinite(config.variables.upper_bounds)):
            upper_values = _get_upper_bound_constraint_values(
                config,
                (
                    evaluations.variables
                    if evaluations.scaled_variables is None
                    else evaluations.scaled_variables
                ),
                axis=-1,
            )
            upper_violations = _get_upper_bound_violations(upper_values)
            unscaled_upper_values = _get_unscaled(upper_values)
            unscaled_upper_violations = _get_unscaled(upper_violations)
            kwargs.update(
                {
                    "upper_values": (
                        upper_values
                        if unscaled_upper_values is None
                        else unscaled_upper_values
                    ),
                    "upper_violations": (
                        upper_violations
                        if unscaled_upper_violations is None
                        else unscaled_upper_violations
                    ),
                    "scaled_upper_values": (
                        None if unscaled_upper_values is None else upper_values
                    ),
                    "scaled_upper_violations": (
                        None if unscaled_upper_violations is None else upper_violations
                    ),
                },
            )
        return BoundConstraints(**kwargs) if kwargs else None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from ropt import utils
from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig


@dataclass
class GradientEvaluations(ResultField):
    """This class contains the results of evaluations for gradient calculation.

    This class stores the variables with the calculated objective and
    constraints gradients. It contains the following information:

    1. The vector of variables at which the functions are evaluated.
    2. A three-dimensional array of perturbed variables, with variable values
       arranged along the third axis. The second axis index indicates the
       perturbation number, whereas the first axis index represents the
       realization number.
    3. The objectives and constraints for each realization and perturbed
       variable vector:  A three-dimensional array, with the objective or
       constraint values arranged along the third axis. The second axis index
       indicates the perturbation number, whereas the first axis index
       represents the realization number.
    4. Scaled versions of the variables and perturbed variables if scaling
       was enabled.
    5. Scaled versions of the objective and constraint values if scaling was
       enabled.
    6. Optional evaluation IDs that may have been passed from the evaluator,
       identifying each calculated realization and perturbation.

    Attributes:
        variables:                    The unperturbed variable vector.
        perturbed_variables:          The variables for each realization and perturbation.
        perturbed_objectives:         The objective functions for each realization and
                                      perturbation.
        perturbed_constraints:        The constraint functions for each realization and
                                      perturbation.
        scaled_variables:             Optional variables after scaling back.
        scaled_perturbed_variables:   Optional variables after scaling back.
        scaled_perturbed_objectives:  Optional scaled objectives.
        scaled_perturbed_constraints: Optional scaled constraints.
        perturbed_evaluation_ids:     Optional id of each evaluated realization.
    """

    variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    perturbed_variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    perturbed_objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.OBJECTIVE,
            ),
        },
    )
    perturbed_constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    scaled_variables: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    scaled_perturbed_variables: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    scaled_perturbed_objectives: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.OBJECTIVE,
            ),
        },
    )
    scaled_perturbed_constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    perturbed_evaluation_ids: Optional[NDArray[np.intc]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
            ),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.variables = _immutable_copy(self.variables)
        self.perturbed_variables = _immutable_copy(self.perturbed_variables)
        self.perturbed_objectives = _immutable_copy(self.perturbed_objectives)
        self.perturbed_constraints = _immutable_copy(self.perturbed_constraints)
        self.scaled_variables = _immutable_copy(self.scaled_variables)
        self.scaled_perturbed_variables = _immutable_copy(
            self.scaled_perturbed_variables
        )
        self.scaled_perturbed_objectives = _immutable_copy(
            self.scaled_perturbed_objectives
        )
        self.scaled_perturbed_constraints = _immutable_copy(
            self.scaled_perturbed_constraints
        )
        self.perturbed_evaluation_ids = _immutable_copy(self.perturbed_evaluation_ids)

    @classmethod
    def create(  # noqa: PLR0913
        cls,
        config: EnOptConfig,
        variables: NDArray[np.float64],
        perturbed_variables: NDArray[np.float64],
        perturbed_objectives: NDArray[np.float64],
        perturbed_constraints: Optional[NDArray[np.float64]] = None,
        objective_auto_scales: Optional[NDArray[np.float64]] = None,
        constraint_auto_scales: Optional[NDArray[np.float64]] = None,
        perturbed_evaluation_ids: Optional[NDArray[np.intc]] = None,
    ) -> GradientEvaluations:
        """Create a FunctionEvaluations object with the given information.

        Args:
            config:                   Configuration object.
            variables:                The unperturbed variable vector.
            perturbed_variables:      The unperturbed variable vector.
            perturbed_objectives:     The objective functions for each realization.
            perturbed_constraints:    The constraint functions for each realization.
            objective_auto_scales:    Objective auto-scaling information.
            constraint_auto_scales:   Constraint auto-scaling information.
            perturbed_evaluation_ids: Optional IDs of the objective calculations.

        Returns:
            A new FunctionEvaluations object.
        """
        unscaled_variables = utils.scaling.scale_back_variables(
            config, variables, axis=-1
        )
        unscaled_perturbed_variables = utils.scaling.scale_back_variables(
            config, perturbed_variables, axis=-1
        )
        scaled_perturbed_objectives = utils.scaling.scale_objectives(
            config,
            perturbed_objectives,
            None if objective_auto_scales is None else objective_auto_scales,
            axis=-1,
        )
        scaled_perturbed_constraints = utils.scaling.scale_constraints(
            config,
            perturbed_constraints,
            None if constraint_auto_scales is None else constraint_auto_scales,
            axis=-1,
        )
        return GradientEvaluations(
            variables=variables if unscaled_variables is None else unscaled_variables,
            perturbed_variables=(
                perturbed_variables
                if unscaled_perturbed_variables is None
                else unscaled_perturbed_variables
            ),
            perturbed_objectives=perturbed_objectives,
            perturbed_constraints=perturbed_constraints,
            scaled_variables=None if unscaled_variables is None else variables,
            scaled_perturbed_variables=(
                None if unscaled_perturbed_variables is None else perturbed_variables
            ),
            scaled_perturbed_objectives=scaled_perturbed_objectives,
            scaled_perturbed_constraints=scaled_perturbed_constraints,
            perturbed_evaluation_ids=perturbed_evaluation_ids,
        )



"""Functions for scaling variables and functions."""

from typing import Optional

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig


def scale_variables(
    config: EnOptConfig, variables: NDArray[np.float64], axis: int
) -> Optional[NDArray[np.float64]]:
    """Scale variables.

    Given an ensemble optimizer configuration object and a vector of variables,
    this function applies an offset and scale to the vector values.

    As the `variables` input may be a multi-dimensional array, the index of the
    axis that designates the variables should be provided through the `axis`
    argument.

    Args:
        config:    The ensemble optimizer configuration object.
        variables: The scaled variables.
        axis:      The variables axis.

    Returns:
        The scaled variables, or `None` if no scaling is applied.
    """
    if config.variables.offsets is None and config.variables.scales is None:
        return None
    variables = np.moveaxis(variables, axis, -1)
    if config.variables.offsets is not None:
        variables = variables - config.variables.offsets
    if config.variables.scales is not None:
        variables = variables / config.variables.scales
    return np.moveaxis(variables, -1, axis)


def scale_back_variables(
    config: EnOptConfig,
    variables: NDArray[np.float64],
    axis: int,
    *,
    correct_offsets: bool = True,
) -> Optional[NDArray[np.float64]]:
    """Scale back variables.

    Given an ensemble optimizer configuration object and a vector of scaled
    variables, scale their values back to the original range. Normally this
    includes correcting for offsets, but if a difference value is being
    rescaled, the `correct_offsets` flag can be used to disable this.

    As the `variables` input may be a multi-dimensional array, the index of the
    axis that designates the variables should be provided through the `axis`
    argument.

    Args:
        config:          The ensemble optimizer configuration object.
        variables:       The scaled variables.
        axis:            The variables axis.
        correct_offsets: If True also correct for offsets.

    Returns:
        The unscaled variables, or `None` if no scaling is applied.
    """
    if config.variables.offsets is None and config.variables.scales is None:
        return None
    variables = np.moveaxis(variables, axis, -1)
    if config.variables.scales is not None:
        variables = variables * config.variables.scales
    if correct_offsets and config.variables.offsets is not None:
        variables = variables + config.variables.offsets
    return np.moveaxis(variables, -1, axis)


def scale_objectives(
    config: EnOptConfig,
    objectives: NDArray[np.float64],
    scales: Optional[NDArray[np.float64]],
    axis: int,
) -> Optional[NDArray[np.float64]]:
    """Scale objective function values.

    Given an ensemble optimizer configuration object, this function scales the
    provided objective values. It divides them by the scale values given in the
    configuration object (if not `None`), and optionally also by the values
    given in the `scales` argument.

    As the `objectives` input may be a multi-dimensional array, the index of the
    axis that designates the objectives should be provided through the `axis`
    argument.

    Args:
        config:     The ensemble optimizer configuration object.
        objectives: Objective functions.
        scales:     Optional additional scales.
        axis:       The objectives axis.

    Returns:
        The scaled objectives or `None` if no scaling was applied.
    """
    total_scales = config.objective_functions.scales
    if scales is not None:
        total_scales = total_scales * scales
    if np.allclose(total_scales, 1.0, rtol=0.0, atol=1e-10):
        return None
    objectives = np.moveaxis(objectives, axis, -1)
    return np.moveaxis(objectives / total_scales, -1, axis)


def scale_constraints(
    config: EnOptConfig,
    constraints: Optional[NDArray[np.float64]],
    scales: Optional[NDArray[np.float64]],
    axis: int,
) -> Optional[NDArray[np.float64]]:
    """Scale constraint function values.

    Given an ensemble optimizer configuration object, this function scales the
    provided constraint values. It divides them by the scale values given in the
    configuration object (if not `None`), and optionally also by the values
    given in the `scales` argument.

    As the `constraints` input may be a multi-dimensional array, the index of the
    axis that designates the constraints should be provided through the `axis`
    argument.

    Args:
        config:      The ensemble optimizer configuration object.
        constraints: Constraint functions.
        scales:      Optional additional scales.
        axis:        The constraints axis.

    Returns:
        The scaled constraints or `None` if no scaling was applied.
    """
    if constraints is None or config.nonlinear_constraints is None:
        return None
    total_scales = config.nonlinear_constraints.scales
    if scales is not None:
        total_scales = total_scales * scales
    if np.allclose(total_scales, 1.0, rtol=0.0, atol=1e-10):
        return None
    constraints = np.moveaxis(constraints, axis, -1)
    return np.moveaxis(constraints / total_scales, -1, axis)



"""This module defines the optimization plan object."""

from __future__ import annotations

import ast
import copy
import re
from numbers import Number
from typing import Any, Callable, Dict, Final, List, Mapping, Optional

import numpy as np

from ropt.config.enopt import EnOptConfig
from ropt.results import Results

_VALID_TYPES: Final = (int, float, bool, str)
_UNARY_OPS: Final = (ast.UAdd, ast.USub, ast.Not)
_BIN_OPS: Final = (ast.Add, ast.Sub, ast.Div, ast.FloorDiv, ast.Mult, ast.Mod, ast.Pow)
_BOOL_OPS: Final = (ast.Or, ast.And)
_CMP_OPS: Final = (ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE)

_SUPPORTED_VARIABLE_TYPES: Final = (
    Number,
    str,
    np.ndarray,
    Dict,
    List,
    Results,
    EnOptConfig,
)
_BUILTIN_FUNCTIONS: Final[Dict[str, Callable[..., Any]]] = {
    "abs": abs,
    "bool": bool,
    "divmod": divmod,
    "float": float,
    "int": int,
    "max": max,
    "min": min,
    "pow": pow,
    "range": range,
    "round": round,
    "sum": sum,
}


class _ReplaceFields(ast.NodeTransformer):
    def __init__(
        self, variables: Mapping[str, Any], functions: Dict[str, Callable[..., Any]]
    ) -> None:
        self.values: Dict[str, Any] = {}
        self._variables = variables
        self._functions = set(functions.keys())

    def visit_Name(self, node: ast.Name) -> ast.AST:  # noqa: N802
        if node.id in self._variables:
            value = self._variables[node.id]
            if value is None or isinstance(value, _SUPPORTED_VARIABLE_TYPES):
                self.values[node.id] = value
                return self.generic_visit(node)
            msg = f"Error in expression: the type of `{node.id}` is not supported"
            raise TypeError(msg)
        if node.id not in self._functions:
            msg = f"Unknown plan variable: `{node.id}`"
            raise AttributeError(msg)
        return self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> ast.AST:  # noqa: N802
        assert isinstance(node.func, ast.Name)
        if node.func.id not in self._functions:
            msg = f"Unknown plan function: `{node.func.id}`"
            raise AttributeError(msg)
        return self.generic_visit(node)


class ExpressionEvaluator:
    """A class for evaluating mathematical expressions in strings."""

    def __init__(self) -> None:
        """Initialize the expression evaluator.

        Raises:
            ValueError: Raised if any provided function overrides a built-in.
        """
        self._functions = copy.deepcopy(_BUILTIN_FUNCTIONS)
        self._functions["_eval"] = self._eval_function

    def add_functions(self, functions: Optional[Dict[str, Callable[..., Any]]]) -> None:
        """Add functions to the evaluator.

        Args:
            functions: A dictionary of functions to add.

        Raises:
            ValueError: If a function already exists.
        """
        if functions is not None:
            for key in functions:
                if key in self._functions:
                    msg = f"cannot override existing function: `{key}`"
                    raise ValueError(msg)
            for key, value in functions.items():
                self._functions[key] = value

    def eval(
        self,
        expr: str,
        variables: Mapping[str, Any],
    ) -> Any:  # noqa: ANN401
        """Evaluate an expression string with provided variable values.

        The `expr` string may take one of the following forms:

        1. Strings enclosed in `$(...)` are evaluated as expressions, that may
           contain variables references and nested calls to functions. Functions
           must be built-in to the expression evaluator, or have been added to
           the evaluator object when it was constructed.

        2. Strings of the form `$name(...)` are evaluated as function calls. The
           contents between the `()` brackets may be an expression that is
           evaluated before passing the result to the function. An expression
           of the form `$name(...)` is equivalent to an expression evaluation of
           the form `$(name(...))`.

        3. Strings starting with `$name`, where name is not a function, i.e. not
           of the form `$name(...)`, are evaluated as variable references. This
           syntax supports nested indexing via `[]` and attribute access with
           `.`. For example, `$var['foo'].bar[0]` is valid if `var` contains a
           compatible structure like a dictionary. Indices within `[]` may
           themselves be expressions that refer to other variables and
           functions. An expression of the form `$name...` is equivalent to an
           expression evaluation of the form `$(name...), where `...` can be any
           combination of indexing and attribute access.

        4. Strings in `[[ ... ]]` are treated as templates. Within these,
           substrings delimited by `<<...>>` are evaluated as if enclosed in
           `$(...)` and the result substituted for the `<<...>>` substring.

        Info: Supported features
            - A subset of Python operators is supported, covering standard
              mathematical, boolean, and comparison operations.

            - Literals include `int`, `float`, `bool`, and `str`. Nested lists
              (`[...]`) and dictionaries (`{...}`) are supported and can reference
              variables or call functions within their content.

            - Allowed types within expressions include numbers, strings, `numpy`
              arrays, dicts, lists, [`Results`][ropt.results.Results], and
              [`EnOptConfig`][ropt.config.enopt.EnOptConfig] objects.

            - Supported built-ins include: `abs`, `bool`, `divmod`, `float`,
              `int`, `max`, `min`, `pow`, `range`, `round`, and `sum`.

        Note: Implicit use of `$(...)` and `[[...]]`
            Where appropriate, the `$(...)` and `[[...]]` delimiters might be
            implicit in some cases.

            For instance, plan steps support an `if` attribute in their
            configuration allowing for conditional execution. This can be
            specified using the `$(...)` syntax, but this is optional. For
            example, instead of `"if": "${x > 0}"`, `"if": "x > 0"` may be used.
            (See [`PlanStepConfig`][ropt.config.plan.PlanStepConfig]).

            Similarly, the [`print`][ropt.plugins.plan._print.DefaultPrintStep]
            step is an example where the `[[...]]` delimiters are optional. This
            step prints a message that will be evaluated, thereby substituting
            any occurrences of `<<...>>`. However, the use of surrounding
            `[[...]]` delimiters in the message attribute of the step
            configuration is optional, and they will be implicitly added if not
            present.

        Args:
            expr:      The expression to evaluate as a string.
            variables: A dictionary of variable values.

        Returns:
            The evaluated result.
        """
        for variable in variables:
            if variable in self._functions:
                msg = f"conflicting variable/function names: `{variable}`"
                raise ValueError(msg)

        expr = expr.strip()
        if expr.startswith("$$"):
            return expr.replace("$$", "$", 1)
        if expr.startswith("[[") and expr.endswith("]]"):
            return self._eval_parts(expr, variables)
        if expr.startswith("$("):
            return self._eval_expr("_eval" + expr[1:], variables)
        if expr.startswith("$"):
            self._check_variable_or_function(expr[1:])
            return self._eval_expr(expr[1:], variables)
        return expr

    def _eval_parts(self, expr: str, variables: Mapping[str, Any]) -> Any:  # noqa: ANN401
        def _substitute(matched: re.Match[str]) -> str:
            parts = matched.string[matched.start() : matched.end()].partition(">>")
            return str(self._eval_expr(parts[0][2:], variables)) + parts[2]

        parts = expr[2:-2].split("<<")
        for part in parts[1:]:
            if part.count(">>") != 1:
                msg = "Missing, or too many, `>>` in [[ ... ]] expression"
                raise SyntaxError(msg)
        parts[1:] = ["<<" + part for part in parts[1:]]
        return "".join(re.sub(r"<<(.*)>>", _substitute, part) for part in parts)

    def _eval_function(self, value: Any) -> Any:  # noqa: ANN401
        return value

    def _is_valid(self, node: ast.AST) -> bool:  # noqa: C901, PLR0911
        if isinstance(node, ast.Constant):
            return node.value is None or type(node.value) in _VALID_TYPES
        if isinstance(node, ast.UnaryOp):
            return isinstance(node.op, _UNARY_OPS) and self._is_valid(node.operand)
        if isinstance(node, ast.BinOp):
            return (
                isinstance(node.op, _BIN_OPS)
                and self._is_valid(node.left)
                and self._is_valid(node.right)
            )
        if isinstance(node, ast.BoolOp):
            return (
                isinstance(node.op, _BOOL_OPS)
                and all(self._is_valid(value) for value in node.values)  # noqa: PD011
            )
        if isinstance(node, ast.Compare):
            return all(isinstance(op, _CMP_OPS) for op in node.ops) and all(
                self._is_valid(value) for value in node.comparators
            )
        if isinstance(node, ast.List):
            return all(self._is_valid(item) for item in node.elts)
        if isinstance(node, ast.Dict):
            return (
                all(item is None or self._is_valid(item) for item in node.keys)
                and all(self._is_valid(item) for item in node.values)  # noqa: PD011
            )
        if isinstance(node, ast.Subscript):
            return self._is_valid(node.slice)
        if isinstance(node, ast.Index):
            return self._is_valid(node.value)  # type: ignore[attr-defined]
        if isinstance(node, ast.Attribute):
            if isinstance(node.value, ast.Attribute):
                return self._is_valid(node.value)
            return bool(isinstance(node.value, ast.Name))
        if isinstance(node, ast.Call):
            return (
                isinstance(node.func, ast.Name)
                and all(self._is_valid(arg) for arg in node.args)
                and not node.keywords
            )
        return bool(isinstance(node, ast.Name))

    def _eval_expr(self, expr: str, variables: Mapping[str, Any]) -> Any:  # noqa: ANN401
        stripped = expr.strip()
        tree = ast.parse(stripped, mode="eval")
        if self._is_valid(tree.body):
            replacer = _ReplaceFields(variables, self._functions)
            tree = ast.fix_missing_locations(replacer.visit(tree))
            return eval(  # noqa: S307
                compile(tree, "", mode="eval"),
                {"__builtins__": self._functions},
                replacer.values,
            )

        msg = "invalid expression"
        raise SyntaxError(msg)

    def _is_valid_variable_or_function(self, node: ast.AST) -> bool:
        if isinstance(node, ast.Subscript):
            return self._is_valid(node.slice)
        if isinstance(node, ast.Index):
            return self._is_valid(node.value)  # type: ignore[attr-defined]
        if isinstance(node, ast.Attribute):
            if isinstance(node.value, ast.Attribute):
                return self._is_valid(node.value)
            return bool(isinstance(node.value, ast.Name))
        if isinstance(node, ast.Call):
            return (
                isinstance(node.func, ast.Name)
                and all(self._is_valid(arg) for arg in node.args)
                and not node.keywords
            )
        return bool(isinstance(node, ast.Name))

    def _check_variable_or_function(self, expr: str) -> Any:  # noqa: ANN401
        stripped = expr.strip()
        tree = ast.parse(stripped, mode="eval")
        if not self._is_valid_variable_or_function(tree.body):
            msg = "invalid variable or function"
            raise SyntaxError(msg)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from ropt import utils
from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig


@dataclass
class Gradients(ResultField):
    """This class stores the calculated objective and constraint gradients.

    The objective and constraint gradients used by the optimizer are calculated
    from their values evaluated for all realizations, for instance by averaging.
    There may be multiple objectives and constraints. Multiple objectives are
    handled by the optimizer using a weighted sum of the individual gradients,
    stored in the `weighted_objective` field. Multiple constraints are directly
    handled by the optimizer.

    If scaling of objectives and/or constraint functions is enabled, also scaled
    versions of the objective or constraint gradients are stored.

    Attributes:
        weighted_objective: The weighted sum of the objective gradients.
        objectives:         The value of each objective gradient.
        constraints:        The value of each constraint gradient.
        scaled_objectives:  Optional scaled objective gradients.
        scaled_constraints: Optional scaled constraint gradients.
    """

    weighted_objective: NDArray[np.float64] = field(
        metadata={"__axes__": (ResultAxisName.VARIABLE,)},
    )
    objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxisName.OBJECTIVE,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.NONLINEAR_CONSTRAINT,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    scaled_objectives: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.OBJECTIVE,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    scaled_constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.NONLINEAR_CONSTRAINT,
                ResultAxisName.VARIABLE,
            ),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.weighted_objective = _immutable_copy(self.weighted_objective)
        self.objectives = _immutable_copy(self.objectives)
        self.constraints = _immutable_copy(self.constraints)
        self.scaled_objectives = _immutable_copy(self.scaled_objectives)
        self.scaled_constraints = _immutable_copy(self.scaled_constraints)

    @classmethod
    def create(  # noqa: PLR0913
        cls,
        config: EnOptConfig,
        weighted_objective: NDArray[np.float64],
        objectives: NDArray[np.float64],
        constraints: Optional[NDArray[np.float64]] = None,
        objective_auto_scales: Optional[NDArray[np.float64]] = None,
        constraint_auto_scales: Optional[NDArray[np.float64]] = None,
    ) -> Gradients:
        """Create a Gradients object with the given information.

        Args:
            config:                 Configuration object.
            weighted_objective:     The weighted objective.
            objectives:             The objective gradients for each realization.
            constraints:            The constraint gradients for each realization.
            objective_auto_scales:  Objective auto-scaling information.
            constraint_auto_scales: Constraint auto-scaling information.

        Returns:
            A new Functions object.
        """
        scaled_objectives = utils.scaling.scale_objectives(
            config,
            objectives,
            None if objective_auto_scales is None else objective_auto_scales,
            axis=0,
        )
        scaled_constraints = utils.scaling.scale_constraints(
            config,
            constraints,
            None if constraint_auto_scales is None else constraint_auto_scales,
            axis=0,
        )
        return Gradients(
            weighted_objective=weighted_objective,
            objectives=objectives,
            constraints=constraints,
            scaled_objectives=scaled_objectives,
            scaled_constraints=scaled_constraints,
        )



"""Plugin functionality for adding function transforms.

This package contains the abstract base classes for function transform plugins,
and the default function transforms that are part of `ropt`.
"""



"""This module implements the SciPy optimization plugin."""

import copy
import os
import sys
from contextlib import redirect_stdout
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Final,
    List,
    Optional,
    Set,
    TextIO,
    Tuple,
    Union,
)

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import (
    Bounds,
    LinearConstraint,
    NonlinearConstraint,
    differential_evolution,
    minimize,
)

from ropt.config.enopt import EnOptConfig
from ropt.enums import ConstraintType, VariableType
from ropt.plugins.optimizer.utils import (
    create_output_path,
    filter_linear_constraints,
    validate_supported_constraints,
)

from .base import Optimizer, OptimizerCallback, OptimizerPlugin

_SUPPORTED_METHODS: Final[Set[str]] = {
    name.lower()
    for name in (
        "Nelder-Mead",
        "Powell",
        "CG",
        "BFGS",
        "Newton-CG",
        "L-BFGS-B",
        "TNC",
        "COBYLA",
        "SLSQP",
        "differential_evolution",
    )
}


# Categorize the methods by the types of constraint they support or require.

_CONSTRAINT_REQUIRES_BOUNDS: Final = {
    "differential_evolution",
}
_CONSTRAINT_SUPPORT_BOUNDS: Final = {
    name.lower()
    for name in [
        "Nelder-Mead",
        "Powell",
        "L-BFGS-B",
        "TNC",
        "SLSQP",
        "differential_evolution",
    ]
}
_CONSTRAINT_SUPPORT_LINEAR_EQ: Final = {
    name.lower() for name in ["SLSQP", "differential_evolution"]
}
_CONSTRAINT_SUPPORT_LINEAR_INEQ: Final = {
    name.lower() for name in ["COBYLA", "SLSQP", "differential_evolution"]
}
_CONSTRAINT_SUPPORT_NONLINEAR_EQ: Final = {
    name.lower() for name in ["SLSQP", "differential_evolution"]
}
_CONSTRAINT_SUPPORT_NONLINEAR_INEQ: Final = {
    name.lower() for name in ["COBYLA", "SLSQP", "differential_evolution"]
}

# These methods do not use a gradient:
_NO_GRADIENT: Final = {
    name.lower()
    for name in ["Nelder-Mead", "Powell", "COBYLA", "differential_evolution"]
}

_OUTPUT_FILE: Final = "optimizer_output"

_ConstraintType = Union[
    str,
    Callable[..., float],
    Callable[..., NDArray[np.float64]],
]


class SciPyOptimizer(Optimizer):
    """Plugin class for optimization via SciPy.

    This class implements several optimizers provided by SciPy in the
    [`scipy.optimize`](https://docs.scipy.org/doc/scipy/tutorial/optimize.html)
    package:

    - Nelder-Mead
    - Powell
    - CG
    - BFGS
    - Newton-CG
    - L-BFGS-B
    - TNC
    - COBYLA
    - SLSQP
    - differential_evolution

    The optimizer to use is selected by setting the `method` field in the
    [`optimizer`][ropt.config.enopt.OptimizerConfig] field of
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] to the name of the algorithm.
    Most of these methods support the general options set in the
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object. However, specific
    options that are normally passed as arguments in the SciPy functions can be
    provided via the `options` dictionary in the configuration object. Consult
    the
    [`scipy.optimize`](https://docs.scipy.org/doc/scipy/tutorial/optimize.html)
    manual for details on these options.

    Not all constraints are supported by all optimizers:

    - Bound constraints: Nelder-Mead, L-BFGS-B, SLSQP, TNC,
      differential_evolution
    - Linear constraints: SLSQP, differential_evolution
    - Nonlinear constraints: COBYLA (only inequality), SLSQP,
      differential_evolution

    Info:
        - The Nelder-Mead algorithm only supports bound constraints if SciPy
          version >= 1.7.
        - Some SciPy algorithms that require a Hessian or a Hessian-vector
          product are not supported. These include dogleg, trust-ncg,
          trust-exact, and trust-krylov.
    """

    _supported_constraints: ClassVar[Dict[str, Set[str]]] = {
        "bounds": _CONSTRAINT_SUPPORT_BOUNDS,
        "linear:eq": _CONSTRAINT_SUPPORT_LINEAR_EQ,
        "linear:ineq": _CONSTRAINT_SUPPORT_LINEAR_INEQ,
        "nonlinear:eq": _CONSTRAINT_SUPPORT_NONLINEAR_EQ,
        "nonlinear:ineq": _CONSTRAINT_SUPPORT_NONLINEAR_INEQ,
    }
    _required_constraints: ClassVar[Dict[str, Set[str]]] = {
        "bounds": _CONSTRAINT_REQUIRES_BOUNDS,
    }

    def __init__(
        self, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> None:
        """Initialize the optimizer implemented by the SciPy plugin.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        self._optimizer_callback = optimizer_callback
        self._config = config
        _, _, self._method = self._config.optimizer.method.lower().rpartition("/")
        if self._method == "default":
            self._method = "slsqp"
        if self._method not in _SUPPORTED_METHODS:
            msg = f"SciPy optimizer algorithm {self._method} is not supported"
            raise NotImplementedError(msg)
        validate_supported_constraints(
            self._config,
            self._method,
            self._supported_constraints,
            self._required_constraints,
        )
        self._bounds = self._initialize_bounds()

        self._constraints: Union[
            Tuple[LinearConstraint, ...],
            List[Dict[str, _ConstraintType]],
        ]

        if self._method == "differential_evolution":
            self._constraints = (
                self._initialize_linear_constraint_object()
                + self._initialize_nonlinear_constraint_object()
            )
        else:
            self._constraints = (
                self._initialize_linear_constraints()
                + self._initialize_nonlinear_constraints()
            )
        self._options = self._parse_options()
        self._parallel = (
            self._config.optimizer.parallel and self._method == "differential_evolution"
        )

        self._cached_variables: Optional[NDArray[np.float64]] = None
        self._cached_function: Optional[NDArray[np.float64]] = None
        self._cached_gradient: Optional[NDArray[np.float64]] = None
        self._stdout: TextIO

    def start(self, initial_values: NDArray[np.float64]) -> None:
        """Start the optimization.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        self._cached_variables = None
        self._cached_function = None
        self._cached_gradient = None

        variable_indices = self._config.variables.indices
        if variable_indices is not None:
            initial_values = np.take(initial_values, variable_indices, axis=-1)

        output_dir = self._config.optimizer.output_dir
        output_file: Union[str, Path]
        if output_dir is None:
            output_file = os.devnull
        else:
            output_file = create_output_path(_OUTPUT_FILE, output_dir, suffix=".txt")

        self._stdout = sys.stdout
        with Path(output_file).open("a", encoding="utf-8") as output, redirect_stdout(
            output,
        ):
            if self._method == "differential_evolution":
                if self._parallel:
                    self._options["updating"] = "deferred"
                    self._options["workers"] = 1
                differential_evolution(
                    func=self._function,
                    x0=initial_values,
                    bounds=self._bounds,
                    constraints=self._constraints,
                    polish=False,
                    vectorized=self._parallel,
                    **self._options,
                )
            else:
                minimize(
                    fun=self._function,
                    x0=initial_values,
                    tol=self._config.optimizer.tolerance,
                    method=self._method,
                    bounds=self._bounds,
                    jac=(False if self._method in _NO_GRADIENT else self._gradient),
                    constraints=self._constraints,
                    options=self._options if self._options else None,
                )

    @property
    def allow_nan(self) -> bool:
        """Whether NaN is allowed.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        return self._method == "differential_evolution"

    @property
    def is_parallel(self) -> bool:
        """Whether the current run is parallel.

        See the [ropt.plugins.optimizer.base.Optimizer][] abstract base class.

        # noqa
        """
        return self._parallel

    def _initialize_bounds(self) -> Optional[Bounds]:
        if (
            np.isfinite(self._config.variables.lower_bounds).any()
            or np.isfinite(self._config.variables.upper_bounds).any()
        ):
            lower_bounds = self._config.variables.lower_bounds
            upper_bounds = self._config.variables.upper_bounds
            variable_indices = self._config.variables.indices
            if variable_indices is not None:
                lower_bounds = lower_bounds[variable_indices]
                upper_bounds = upper_bounds[variable_indices]
            return Bounds(lower_bounds, upper_bounds)
        return None

    def _initialize_linear_constraints(self) -> List[Dict[str, _ConstraintType]]:
        constraints: List[Dict[str, _ConstraintType]] = []

        linear_constraints_config = self._config.linear_constraints
        if linear_constraints_config is None:
            return constraints

        if self._config.variables.indices is not None:
            linear_constraints_config = filter_linear_constraints(
                linear_constraints_config, self._config.variables.indices
            )

        types = linear_constraints_config.types
        coefficients = linear_constraints_config.coefficients
        rhs_values = linear_constraints_config.rhs_values

        eq_idx = types == ConstraintType.EQ
        if np.any(eq_idx):
            assert coefficients is not None
            constraints.append(
                {
                    "type": "eq",
                    "fun": lambda x: (
                        np.matmul(coefficients[eq_idx, :], x) - rhs_values[eq_idx]
                    ),
                    "jac": lambda _: coefficients[eq_idx, :],
                },
            )

        ineq_constraints = []
        ineq_rhs_values = []
        ge_idx = types == ConstraintType.GE
        if np.any(ge_idx):
            assert coefficients is not None
            assert rhs_values is not None
            ineq_constraints.append(coefficients[ge_idx, :])
            ineq_rhs_values.append(rhs_values[ge_idx])
        le_idx = types == ConstraintType.LE
        if np.any(le_idx):
            assert coefficients is not None
            assert rhs_values is not None
            ineq_constraints.append(-coefficients[le_idx, :])
            ineq_rhs_values.append(-rhs_values[le_idx])
        if ineq_constraints:
            coefficients_matrix = np.vstack(tuple(ineq_constraints))
            rhs_vector = np.hstack(tuple(ineq_rhs_values))
            constraints.append(
                {
                    "type": "ineq",
                    "fun": lambda x: (np.matmul(coefficients_matrix, x) - rhs_vector),
                    "jac": lambda _: coefficients_matrix,
                },
            )

        return constraints

    def _initialize_linear_constraint_object(self) -> Tuple[LinearConstraint, ...]:
        linear_constraints_config = self._config.linear_constraints
        if linear_constraints_config is None:
            return ()

        if self._config.variables.indices is not None:
            linear_constraints_config = filter_linear_constraints(
                linear_constraints_config, self._config.variables.indices
            )

        lower_bounds = linear_constraints_config.rhs_values.copy()
        upper_bounds = linear_constraints_config.rhs_values.copy()

        ge_idx = linear_constraints_config.types == ConstraintType.GE
        le_idx = linear_constraints_config.types == ConstraintType.LE
        if np.any(le_idx):
            lower_bounds[le_idx] = -np.inf
        if np.any(ge_idx):
            upper_bounds[ge_idx] = np.inf

        return (
            LinearConstraint(
                linear_constraints_config.coefficients, lower_bounds, upper_bounds
            ),
        )

    def _initialize_nonlinear_constraints(self) -> List[Dict[str, _ConstraintType]]:
        constraints: List[Dict[str, _ConstraintType]] = []

        if self._config.nonlinear_constraints is None:
            return constraints

        for inx, constraint_type in enumerate(self._config.nonlinear_constraints.types):
            constr = "eq" if constraint_type == ConstraintType.EQ else "ineq"
            fun = partial(self._constraint_function, index=inx)
            if self._method == "cobyla":
                constraints.append({"type": constr, "fun": fun})
            else:
                jac = partial(self._constraint_gradient, index=inx)
                constraints.append({"type": constr, "fun": fun, "jac": jac})
        return constraints

    def _initialize_nonlinear_constraint_object(
        self,
    ) -> Tuple[NonlinearConstraint, ...]:
        if self._config.nonlinear_constraints is None:
            return ()

        constraints_count = self._config.nonlinear_constraints.rhs_values.size

        def fun(variables: NDArray[np.float64]) -> NDArray[np.float64]:
            if variables.ndim == 1:
                result = np.empty(constraints_count, dtype=np.float64)
                for inx in range(constraints_count):
                    result[inx] = partial(self._constraint_function, index=inx)(
                        variables
                    )
            else:
                result = np.empty(
                    (constraints_count, variables.shape[1]), dtype=np.float64
                )
                for inx in range(constraints_count):
                    result[inx, :] = partial(self._constraint_function, index=inx)(
                        variables
                    )
            return result

        def jac(variables: NDArray[np.float64]) -> NDArray[np.float64]:
            result = np.empty((constraints_count, variables.size), dtype=np.float64)
            for inx in range(constraints_count):
                result[inx, :] = partial(self._constraint_gradient, index=inx)(
                    variables
                )
            return result

        return (NonlinearConstraint(fun=fun, jac=jac, lb=-np.inf, ub=0.0),)

    def _function(self, variables: NDArray[np.float64]) -> NDArray[np.float64]:
        if variables.ndim > 1 and variables.size == 0:
            return np.array([])
        functions, _ = self._get_function_or_gradient(
            variables, get_function=True, get_gradient=False
        )
        assert functions is not None
        if variables.ndim > 1:
            return functions[:, 0]
        return np.array(functions[0])

    def _gradient(self, variables: NDArray[np.float64]) -> NDArray[np.float64]:
        _, gradients = self._get_function_or_gradient(
            variables, get_function=False, get_gradient=True
        )
        assert gradients is not None
        return gradients[0, :]

    def _constraint_function(
        self, variables: NDArray[np.float64], index: int
    ) -> NDArray[np.float64]:
        if variables.ndim > 1 and variables.size == 0:
            return np.array([])
        functions, _ = self._get_function_or_gradient(
            variables, get_function=True, get_gradient=False
        )
        assert functions is not None
        if variables.ndim > 1:
            return -functions[:, index + 1]
        return np.array(-functions[index + 1])

    def _constraint_gradient(
        self, variables: NDArray[np.float64], index: int
    ) -> NDArray[np.float64]:
        _, gradients = self._get_function_or_gradient(
            variables, get_function=False, get_gradient=True
        )
        assert gradients is not None
        return -gradients[index + 1, :]

    def _get_function_or_gradient(
        self, variables: NDArray[np.float64], *, get_function: bool, get_gradient: bool
    ) -> Tuple[Optional[NDArray[np.float64]], Optional[NDArray[np.float64]]]:
        if self._parallel and variables.ndim > 1:
            variables = variables.T

        if self._method in _NO_GRADIENT:
            get_gradient = False

        if (
            self._cached_variables is None
            or variables.shape != self._cached_variables.shape
            or not np.allclose(variables, self._cached_variables)
        ):
            self._cached_variables = None
            self._cached_function = None
            self._cached_gradient = None

        function = self._cached_function if get_function else None
        gradient = self._cached_gradient if get_gradient else None

        compute_functions = get_function and function is None
        compute_gradients = get_gradient and gradient is None

        if compute_functions or compute_gradients:
            self._cached_variables = variables.copy()
            compute_functions = compute_functions or self._config.optimizer.speculative
            compute_gradients = compute_gradients or self._config.optimizer.speculative
            new_function, new_gradient = self._compute_functions_and_gradients(
                variables,
                compute_functions=compute_functions,
                compute_gradients=compute_gradients,
            )
            if compute_functions:
                assert new_function is not None
                self._cached_function = new_function.copy()
                if get_function:
                    function = new_function
            if compute_gradients:
                assert new_gradient is not None
                self._cached_gradient = new_gradient.copy()
                if get_gradient:
                    gradient = new_gradient

        return function, gradient

    def _compute_functions_and_gradients(
        self,
        variables: NDArray[np.float64],
        *,
        compute_functions: bool,
        compute_gradients: bool,
    ) -> Tuple[Optional[NDArray[np.float64]], Optional[NDArray[np.float64]]]:
        new_function = None
        new_gradient = None
        if (
            compute_functions
            and compute_gradients
            and self._config.optimizer.split_evaluations
        ):
            with redirect_stdout(self._stdout):
                new_function, _ = self._optimizer_callback(
                    variables,
                    return_functions=True,
                    return_gradients=False,
                )
                _, new_gradient = self._optimizer_callback(
                    variables,
                    return_functions=False,
                    return_gradients=True,
                )
        else:
            with redirect_stdout(self._stdout):
                new_function, new_gradient = self._optimizer_callback(
                    variables,
                    return_functions=compute_functions,
                    return_gradients=compute_gradients,
                )
        if self.allow_nan:
            new_function = np.where(np.isnan(new_function), np.inf, new_function)
        return new_function, new_gradient

    def _parse_options(self) -> Dict[str, Any]:
        if not isinstance(self._config.optimizer.options, dict):
            return {}
        options = copy.deepcopy(self._config.optimizer.options)
        # The maximum number of iterations is passed as an option to ropt.
        # Setting maxiter directly as an entry in the options dict will also
        # work, but iterations will override it.
        iterations = self._config.optimizer.max_iterations
        if iterations is not None:
            if self._method == "tnc":
                options["maxfun"] = iterations
            else:
                options["maxiter"] = iterations
        # We switch on display if there is an output folder.
        if self._config.optimizer.output_dir is not None:
            options["disp"] = True

        if (
            self._method == "differential_evolution"
            and self._config.variables.types is not None
            and "integrality" not in options
        ):
            options["integrality"] = (
                self._config.variables.types == VariableType.INTEGER
            )

        return options


class SciPyOptimizerPlugin(OptimizerPlugin):
    """Default filter transform plugin class."""

    def create(
        self, config: EnOptConfig, optimizer_callback: OptimizerCallback
    ) -> SciPyOptimizer:
        """Initialize the optimizer plugin.

        See the [ropt.plugins.optimizer.base.OptimizerPlugin][] abstract base class.

        # noqa
        """
        return SciPyOptimizer(config, optimizer_callback)

    def is_supported(self, method: str, *, explicit: bool) -> bool:  # noqa: ARG002
        """Check if a method is supported.

        See the [ropt.plugins.base.Plugin][] abstract base class.

        # noqa
        """
        return method.lower() in (_SUPPORTED_METHODS | {"default"})



"""This module implements the save step."""

from __future__ import annotations

import json
import pickle
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Union

from pydantic import BaseModel, ConfigDict

from ropt.config.validated_types import ItemOrTuple  # noqa: TCH001
from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.config.plan import PlanStepConfig
    from ropt.plan import Plan


class DefaultSaveStep(PlanStep):
    """The default save step.

    The save step saves specified plan variables to a file by
    constructing a dictionary that maps each variable name to its value.
    This dictionary is then serialized and saved in the specified file
    format.

    This step uses the
    [`DefaultSaveStepWith`][ropt.plugins.plan._save_vars.DefaultSaveStep.DefaultSaveStepWith]
    configuration class to parse the `with` field of the
    [`PlanStepConfig`][ropt.config.plan.PlanStepConfig]. The configuration
    specifies the settings for this step in a plan setup.
    """

    class DefaultSaveStepWith(BaseModel):
        """Configuration parameters for the save step.

        This configuration defines which plan variables are saved by the save
        step and specifies the file path for storing the output. If the path
        includes directories that do not yet exist, they will be created.

        The `format` option defines the file format for saving the data.
        Supported formats are:

        - `json`:   Saves the variables as a dictionary in JSON format.
        - `pickle`: Saves the variables as a dictionary in a pickle file.

        Attributes:
            vars:   List of variable names to save to the specified output file.
            path:   The file path where the output will be stored.
            format: The format used for file storage, determining the serialization method.
        """

        vars: ItemOrTuple[str]
        path: Union[str, Path]
        format: Literal["json", "pickle"] = "json"

        model_config = ConfigDict(
            extra="forbid",
            validate_default=True,
            arbitrary_types_allowed=True,
            frozen=True,
        )

    def __init__(self, config: PlanStepConfig, plan: Plan) -> None:
        """Initialize a default save step.

        Args:
            config: The configuration of the step.
            plan:   The plan that runs this step.
        """
        super().__init__(config, plan)
        self._with = self.DefaultSaveStepWith.model_validate(config.with_)
        for name in self._with.vars:
            if name not in self.plan:
                msg = f"Plan variable does not exist: {name}"
                raise ValueError(msg)
        if self._with.format not in {"json", "pickle"}:
            msg = f"data format not supported: {self._with.format}"
            raise ValueError(msg)

    def run(self) -> None:
        """Run the save step."""
        path = Path(self.plan.eval(self._with.path))
        if path.parent.exists() and not path.parent.is_dir():
            msg = f"Not a directory to store results: {path.parent}"
            raise RuntimeError(msg)
        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)

        data = {name: self.plan[name] for name in self._with.vars}
        if self._with.format == "json":
            with path.open("w", encoding="utf-8") as file_obj:
                json.dump(data, file_obj)
        elif self._with.format == "pickle":
            with path.open("wb") as file_obj:
                pickle.dump(data, file_obj)



from dataclasses import InitVar, dataclass
from itertools import zip_longest
from typing import Generator, Optional, Tuple

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig
from ropt.evaluator import Evaluator, EvaluatorContext


@dataclass
class _FunctionEvaluatorResults:
    batch_id: Optional[int]
    objectives: NDArray[np.float64]
    constraints: Optional[NDArray[np.float64]]
    evaluation_ids: Optional[NDArray[np.intc]]

    def __post_init__(self) -> None:
        self.objectives, self.constraints = _propagate_nan_values(
            self.objectives, self.constraints
        )


@dataclass
class _GradientEvaluatorResults:
    batch_id: Optional[int]
    perturbed_objectives: NDArray[np.float64]
    perturbed_constraints: Optional[NDArray[np.float64]]
    perturbed_evaluation_ids: Optional[NDArray[np.intc]]
    realization_count: InitVar[int]
    perturbation_count: InitVar[int]

    def __post_init__(self, realization_count: int, perturbation_count: int) -> None:
        self.perturbed_objectives, self.perturbed_constraints = _propagate_nan_values(
            self.perturbed_objectives, self.perturbed_constraints
        )

        shape = (realization_count, perturbation_count, -1)
        self.perturbed_objectives = self.perturbed_objectives.reshape(shape)
        self.perturbed_constraints = (
            None
            if self.perturbed_constraints is None
            else self.perturbed_constraints.reshape(shape)
        )
        self.perturbed_evaluation_ids = (
            None
            if self.perturbed_evaluation_ids is None
            else self.perturbed_evaluation_ids.reshape(shape[:2])
        )


def _propagate_nan_values(
    objective_results: NDArray[np.float64],
    constraint_results: Optional[NDArray[np.float64]],
) -> Tuple[NDArray[np.float64], Optional[NDArray[np.float64]]]:
    failures = None
    if objective_results is not None:
        failures = np.logical_or.reduce(np.isnan(objective_results), axis=-1)
    if constraint_results is not None:
        constraint_failures = np.logical_or.reduce(
            np.isnan(constraint_results), axis=-1
        )
        if failures is None:
            failures = constraint_failures
        else:
            failures |= constraint_failures
    if objective_results is not None:
        objective_results = objective_results.copy()
        objective_results[failures, :] = np.nan
    if constraint_results is not None:
        constraint_failures = constraint_failures.copy()
        constraint_results[failures, :] = np.nan
    return objective_results, constraint_results


def _get_active_realizations(
    config: EnOptConfig,
    *,
    objective_weights: Optional[NDArray[np.float64]] = None,
    constraint_weights: Optional[NDArray[np.float64]] = None,
) -> Tuple[Optional[NDArray[np.bool_]], Optional[NDArray[np.bool_]]]:
    if objective_weights is None:
        active_realizations = np.abs(config.realizations.weights) > 0
        if np.all(active_realizations):
            return None, None
        active_objectives = np.broadcast_to(
            active_realizations,
            (config.objective_functions.weights.size, active_realizations.size),
        )
        active_constraints = (
            None
            if config.nonlinear_constraints is None
            else np.broadcast_to(
                active_realizations,
                (
                    config.nonlinear_constraints.rhs_values.size,
                    active_realizations.size,
                ),
            )
        )
        return active_objectives, active_constraints
    active_objectives = np.abs(objective_weights) > 0
    active_constraints = (
        None if constraint_weights is None else np.abs(constraint_weights) > 0
    )
    return active_objectives, active_constraints


def _get_function_results(
    config: EnOptConfig,
    evaluator: Evaluator,
    variables: NDArray[np.float64],
    active_objectives: Optional[NDArray[np.bool_]],
    active_constraints: Optional[NDArray[np.bool_]],
) -> Generator[Tuple[int, _FunctionEvaluatorResults], None, None]:
    realization_num = config.realizations.weights.size
    context = EvaluatorContext(
        config=config,
        realizations=np.tile(
            np.arange(realization_num, dtype=np.intc), variables.shape[0]
        ),
        active_objectives=active_objectives,
        active_constraints=active_constraints,
    )
    evaluator_result = evaluator(np.repeat(variables, realization_num, axis=0), context)
    split_objectives = np.vsplit(evaluator_result.objectives, variables.shape[0])
    split_constraints = (
        []
        if evaluator_result.constraints is None
        else np.vsplit(evaluator_result.constraints, variables.shape[0])
    )
    split_evaluation_ids = (
        []
        if evaluator_result.evaluation_ids is None
        else np.split(evaluator_result.evaluation_ids, variables.shape[0])
    )
    for idx, (objectives, constraints, evaluation_ids) in enumerate(
        zip_longest(
            split_objectives,
            split_constraints,
            split_evaluation_ids,
        )
    ):
        yield (
            idx,
            _FunctionEvaluatorResults(
                batch_id=evaluator_result.batch_id,
                objectives=objectives,
                constraints=constraints,
                evaluation_ids=evaluation_ids,
            ),
        )


def _get_gradient_results(
    config: EnOptConfig,
    evaluator: Evaluator,
    perturbed_variables: NDArray[np.float64],
    active_objectives: Optional[NDArray[np.bool_]],
    active_constraints: Optional[NDArray[np.bool_]],
) -> _GradientEvaluatorResults:
    realization_num = config.realizations.weights.size
    perturbation_num = config.gradient.number_of_perturbations
    context = EvaluatorContext(
        config=config,
        realizations=np.repeat(np.arange(realization_num), perturbation_num),
        active_objectives=active_objectives,
        active_constraints=active_constraints,
    )
    evaluator_result = evaluator(
        perturbed_variables.reshape(-1, perturbed_variables.shape[-1]), context
    )
    return _GradientEvaluatorResults(
        batch_id=evaluator_result.batch_id,
        perturbed_objectives=evaluator_result.objectives,
        perturbed_constraints=evaluator_result.constraints,
        perturbed_evaluation_ids=evaluator_result.evaluation_ids,
        realization_count=config.realizations.weights.size,
        perturbation_count=config.gradient.number_of_perturbations,
    )


def _get_function_and_gradient_results(  # noqa: PLR0913
    config: EnOptConfig,
    evaluator: Evaluator,
    variables: NDArray[np.float64],
    perturbed_variables: NDArray[np.float64],
    active_objectives: Optional[NDArray[np.bool_]],
    active_constraints: Optional[NDArray[np.bool_]],
) -> Tuple[_FunctionEvaluatorResults, _GradientEvaluatorResults]:
    realization_num = config.realizations.weights.size
    perturbation_num = config.gradient.number_of_perturbations
    realizations = np.arange(realization_num)
    context = EvaluatorContext(
        config=config,
        realizations=np.hstack(
            (realizations, np.repeat(realizations, perturbation_num)),
        ),
        active_objectives=active_objectives,
        active_constraints=active_constraints,
    )
    evaluator_result = evaluator(
        np.vstack(
            (
                np.repeat(variables[np.newaxis, ...], realization_num, axis=0),
                perturbed_variables.reshape(-1, perturbed_variables.shape[-1]),
            ),
        ),
        context,
    )
    return (
        _FunctionEvaluatorResults(
            batch_id=evaluator_result.batch_id,
            objectives=evaluator_result.objectives[:realization_num],
            constraints=(
                None
                if evaluator_result.constraints is None
                else evaluator_result.constraints[:realization_num]
            ),
            evaluation_ids=(
                None
                if evaluator_result.evaluation_ids is None
                else evaluator_result.evaluation_ids[:realization_num]
            ),
        ),
        _GradientEvaluatorResults(
            batch_id=evaluator_result.batch_id,
            perturbed_objectives=evaluator_result.objectives[realization_num:, :],
            perturbed_constraints=(
                None
                if evaluator_result.constraints is None
                else evaluator_result.constraints[realization_num:, :]
            ),
            perturbed_evaluation_ids=(
                None
                if evaluator_result.evaluation_ids is None
                else evaluator_result.evaluation_ids[realization_num:]
            ),
            realization_count=config.realizations.weights.size,
            perturbation_count=config.gradient.number_of_perturbations,
        ),
    )



from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, List, Sequence, Tuple

import numpy as np
import pytest
from numpy.typing import NDArray

from ropt.evaluator import Evaluator, EvaluatorContext, EvaluatorResult
from ropt.utils.scaling import scale_back_variables

_Function = Callable[[NDArray[np.float64], Any], float]


def pytest_addoption(parser: Any) -> Any:
    parser.addoption(
        "--run-slow", action="store_true", default=False, help="run slow tests"
    )


def pytest_collection_modifyitems(config: Any, items: Sequence[Any]) -> None:
    if not config.getoption("--run-slow"):
        skip_slow = pytest.mark.skip(reason="need --run-slow option to run")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)


@dataclass
class FunctionContext:
    realization: int
    index: int


def _function_runner(
    variables: NDArray[np.float64],
    evaluator_context: EvaluatorContext,
    functions: List[_Function],
) -> EvaluatorResult:
    unscaled_variables = scale_back_variables(
        evaluator_context.config, variables, axis=-1
    )
    if unscaled_variables is not None:
        variables = unscaled_variables
    objective_count = evaluator_context.config.objective_functions.weights.size
    constraint_count = (
        0
        if evaluator_context.config.nonlinear_constraints is None
        else evaluator_context.config.nonlinear_constraints.rhs_values.size
    )
    objective_results = np.zeros(
        (variables.shape[0], objective_count),
        dtype=np.float64,
    )
    constraint_results = (
        np.zeros((variables.shape[0], constraint_count), dtype=np.float64)
        if constraint_count > 0
        else None
    )
    for sim, realization in enumerate(evaluator_context.realizations):
        context = FunctionContext(realization=realization, index=sim)
        for idx in range(objective_count):
            if (
                evaluator_context.active_objectives is None
                or evaluator_context.active_objectives[idx, realization]
            ):
                function = functions[idx]
                objective_results[sim, idx] = function(variables[sim, :], context)
        for idx in range(constraint_count):
            if (
                evaluator_context.active_constraints is None
                or evaluator_context.active_constraints[idx, realization]
            ):
                function = functions[idx + objective_count]
                assert constraint_results is not None
                constraint_results[sim, idx] = function(variables[sim, :], context)
    return EvaluatorResult(
        objectives=objective_results,
        constraints=constraint_results,
    )


def _compute_distance_squared(
    variables: NDArray[np.float64],
    context: Any,  # noqa: ARG001
    target: NDArray[np.float64],
) -> float:
    return float(((variables - target) ** 2).sum())


@pytest.fixture(name="test_functions", scope="session")
def fixture_test_functions() -> Tuple[_Function, _Function]:
    return (
        partial(_compute_distance_squared, target=np.array([0.5, 0.5, 0.5])),
        partial(_compute_distance_squared, target=np.array([-1.5, -1.5, 0.5])),
    )


@pytest.fixture(scope="session")
def evaluator(test_functions: Any) -> Callable[[List[_Function]], Evaluator]:
    def _evaluator(test_functions: List[_Function] = test_functions) -> Evaluator:
        return partial(_function_runner, functions=test_functions)

    return _evaluator


@pytest.fixture(scope="session")
def assert_equal_dicts() -> Callable[[Any, Any], None]:
    def _assert_equal_dicts(value1: Any, value2: Any) -> None:
        if isinstance(value1, dict):
            assert isinstance(value2, dict)
            for key, item1 in value1.items():
                assert key in value2
                _assert_equal_dicts(item1, value2[key])
        elif isinstance(value1, list):
            assert isinstance(value2, list)
            for item1, item2 in zip(value1, value2):
                _assert_equal_dicts(item1, item2)
        elif isinstance(value1, tuple):
            assert isinstance(value2, tuple)
            for item1, item2 in zip(value1, value2):
                _assert_equal_dicts(item1, item2)
        elif isinstance(value1, np.ndarray):
            assert isinstance(value2, np.ndarray)
            assert np.allclose(value1, value2)
        else:
            assert value1 == value2

    return _assert_equal_dicts



"""This module implements the default save handler."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Sequence, Union

from pydantic import BaseModel, ConfigDict

from ropt.config.validated_types import ItemOrSet  # noqa: TCH001
from ropt.enums import EventType
from ropt.plugins.plan.base import ResultHandler
from ropt.utils.misc import format_tuple

if TYPE_CHECKING:
    from ropt.config.plan import ResultHandlerConfig
    from ropt.plan import Event, Plan


class _Formatter:
    def __init__(self, value: Sequence[Any], delimiters: str) -> None:
        self._value = value
        self._delimiters = delimiters

    def __format__(self, spec: str) -> str:
        return format_tuple(
            tuple(f"{item:{spec}}" for item in self._value), delimiters=self._delimiters
        )


class DefaultSaveHandler(ResultHandler):
    """The default save handler object.

    This handler tracks the [`Results`][ropt.results.Results] objects that it
    receives saves them. It uses the [`DefaultSaveHandlerWith`]
    [ropt.plugins.plan._save_results.DefaultSaveHandler.DefaultSaveHandlerWith]
    configuration class to parse the `with` field of the
    [`ResultHandler`][ropt.config.plan.ResultHandlerConfig] used to specify this
    handler in a plan configuration.
    """

    class DefaultSaveHandlerWith(BaseModel):
        """Parameters for the save handler.

        The `path` field specifies the output path, including the file name. It
        may be a `Path` object, or a string. In the latter case, it may contain
        one or more replacement strings for the following variables, which will
        be substituted with the corresponding value of the result objects:

        - `plan_id`:   Replace with a formatted `plan_id` field of the result object.
        - `result_id`: Replace with the `result_id` field of the result object.
        - `batch_id`:  Replace with the `batch_id` field of the result object.

        For example the following string will construct file names using the
        result ID padded with zeros: "output_directory/result{result_id:03d}.nc".

        Since the `plan_id` field of a results objects is a tuple, it needs to
        formatted before substitution. This is done by joining them using
        delimiters from the `delimiters` attribute. The `delimiters` argument
        can contain multiple characters, each of which will be used in turn to
        separate the items in `name`. If fewer delimiters are provided than
        needed, the final delimiter is reused for any remaining items. If
        `delimiters` is not specified, `-` is used by default.

        The `tags` field allows optional labels to be attached to each result,
        assisting result handlers in filtering relevant results.

        Currently the output is stored in netCDf format. A `.nc` extension is added
        to the file name if not present already.

        Attributes:
            path:       The path to the output file names.
            tags:       Tags to filter the sources to track.
            delimiters: The delimiters to use to format plan_id values, applied sequentially.
        """

        path: Union[str, Path]
        tags: ItemOrSet[str]
        delimiters: str = "-"

        model_config = ConfigDict(
            extra="forbid",
            validate_default=True,
            frozen=True,
        )

    def __init__(self, config: ResultHandlerConfig, plan: Plan) -> None:
        """Initialize a default save handler object.

        Args:
            config: The configuration of the step.
            plan:   The plan that runs this step.
        """
        super().__init__(config, plan)
        self._with = self.DefaultSaveHandlerWith.model_validate(config.with_)

    def handle_event(self, event: Event) -> Event:
        """Handle an event.

        Args:
            event: The event to handle.

        Returns:
            The (possibly modified) event.
        """
        if (
            event.event_type
            in {
                EventType.FINISHED_EVALUATION,
                EventType.FINISHED_EVALUATOR_STEP,
            }
            and event.results is not None
            and (event.tags & self._with.tags)
        ):
            for results in event.results:
                if isinstance(self._with.path, Path):
                    path = self._with.path
                else:
                    path = Path(
                        self.plan.eval(
                            self._with.path.format(
                                plan_id=_Formatter(
                                    results.plan_id, self._with.delimiters
                                ),
                                result_id=results.result_id,
                                batch_id=results.batch_id,
                            )
                        )
                    )
                if path.parent.exists() and not path.parent.is_dir():
                    msg = f"Not a directory to store results: {path.parent}"
                    raise RuntimeError(msg)
                if not path.parent.exists():
                    path.parent.mkdir(parents=True, exist_ok=True)
                if path.suffix != ".nc":
                    path = path.with_suffix(".nc")
                results.to_netcdf(event.config, path)
        return event



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from ropt import utils
from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig


@dataclass
class FunctionEvaluations(ResultField):
    """This class contains the results of evaluations for function calculation.

    This class stores the variables with the calculated objective and constraint
    functions. It contains the following information:

    1. The vector of variables at which the functions are evaluated.
    2. The calculated objectives and constraints for each realization: A
       two-dimensional array, with the objective or constraint values arranged
       along the second axis. The first axis index indicates the realization
       number.
    3. Scaled versions of the variables if variable scaling was enabled.
    4. Scaled versions of the objective and constraint values if scaling was
       enabled.
    5. Optional evaluation IDs that may have been passed from the evaluator,
       identifying each calculated realization.

    Attributes:
        variables:          The unperturbed variable vector.
        objectives:         The objective functions for each realization.
        constraints:        The constraint functions for each realization.
        scaled_variables:   Optional variables after scaling.
        scaled_objectives:  Optional scaled objectives.
        scaled_constraints: Optional scaled constraints.
        evaluation_ids:     Optional id of each evaluated realization.
    """

    variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.OBJECTIVE,
            ),
        },
    )
    constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    scaled_variables: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    scaled_objectives: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.OBJECTIVE,
            ),
        },
    )
    scaled_constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    evaluation_ids: Optional[NDArray[np.intc]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.REALIZATION,),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.variables = _immutable_copy(self.variables)
        self.objectives = _immutable_copy(self.objectives)
        self.constraints = _immutable_copy(self.constraints)
        self.scaled_variables = _immutable_copy(self.scaled_variables)
        self.scaled_objectives = _immutable_copy(self.scaled_objectives)
        self.scaled_constraints = _immutable_copy(self.scaled_constraints)
        self.evaluation_ids = _immutable_copy(self.evaluation_ids)

    @classmethod
    def create(  # noqa: PLR0913
        cls,
        config: EnOptConfig,
        variables: NDArray[np.float64],
        objectives: NDArray[np.float64],
        constraints: Optional[NDArray[np.float64]] = None,
        objective_auto_scales: Optional[NDArray[np.float64]] = None,
        constraint_auto_scales: Optional[NDArray[np.float64]] = None,
        evaluation_ids: Optional[NDArray[np.intc]] = None,
    ) -> FunctionEvaluations:
        """Create a FunctionEvaluations object with the given information.

        Args:
            config:                 Configuration object.
            variables:              The unperturbed variable vector.
            objectives:             The objective functions for each realization.
            constraints:            The constraint functions for each realization.
            objective_auto_scales:  Objective auto-scaling information.
            constraint_auto_scales: Constraint auto-scaling information.
            evaluation_ids:         Optional IDs of the objective calculations.

        Returns:
            A new FunctionEvaluations object.
        """
        unscaled_variables = utils.scaling.scale_back_variables(
            config, variables, axis=-1
        )
        scaled_objectives = utils.scaling.scale_objectives(
            config,
            objectives,
            None if objective_auto_scales is None else objective_auto_scales,
            axis=-1,
        )
        scaled_constraints = utils.scaling.scale_constraints(
            config,
            constraints,
            None if constraint_auto_scales is None else constraint_auto_scales,
            axis=-1,
        )
        return FunctionEvaluations(
            variables=variables if unscaled_variables is None else unscaled_variables,
            objectives=objectives,
            constraints=constraints,
            scaled_variables=None if unscaled_variables is None else variables,
            scaled_objectives=scaled_objectives,
            scaled_constraints=scaled_constraints,
            evaluation_ids=evaluation_ids,
        )



from typing import Any, Optional, overload

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig
from ropt.enums import ConstraintType


def _get_lower_bound_constraint_values(
    config: EnOptConfig, variables: NDArray[np.float64], axis: int
) -> NDArray[np.float64]:
    variables = np.moveaxis(variables, axis, -1)
    bounds = config.variables.lower_bounds
    result = np.where(np.isfinite(bounds), variables - bounds, 0.0)
    return np.moveaxis(result, -1, axis)


def _get_lower_bound_violations(
    constraints: NDArray[np.float64],
) -> NDArray[np.float64]:
    return np.maximum(-constraints, 0.0)


def _get_upper_bound_constraint_values(
    config: EnOptConfig, variables: NDArray[np.float64], axis: int
) -> NDArray[np.float64]:
    variables = np.moveaxis(variables, axis, -1)
    bounds = config.variables.upper_bounds
    result = np.where(np.isfinite(bounds), variables - bounds, 0.0)
    return np.moveaxis(result, -1, axis)


def _get_upper_bound_violations(
    constraints: NDArray[np.float64],
) -> NDArray[np.float64]:
    return np.maximum(constraints, 0.0)


def _get_linear_constraint_values(
    config: EnOptConfig, variables: NDArray[np.float64], axis: int
) -> NDArray[np.float64]:
    assert config.linear_constraints is not None
    coefficients = config.linear_constraints.coefficients
    rhs_values = config.linear_constraints.rhs_values
    variables = np.moveaxis(variables, axis, -1)
    constraint_values = np.empty(
        variables.shape[:-1] + (rhs_values.size,), dtype=np.float64
    )
    for idx in np.ndindex(*variables.shape[:-1]):
        constraint_values[idx] = np.matmul(coefficients, variables[idx])
    result = np.array(constraint_values - rhs_values)
    return np.moveaxis(result, -1, axis)


def _get_linear_constraint_violations(
    config: EnOptConfig, constraints: NDArray[np.float64]
) -> NDArray[np.float64]:
    assert config.linear_constraints is not None
    return _get_constraint_violations(constraints, config.linear_constraints.types)


def _get_nonlinear_constraint_values(
    config: EnOptConfig, constraints: NDArray[np.float64], axis: int
) -> NDArray[np.float64]:
    constraints = np.moveaxis(constraints, axis, -1)
    assert config.nonlinear_constraints is not None
    rhs_values = config.nonlinear_constraints.rhs_values
    assert constraints is not None
    result = constraints - rhs_values
    return np.moveaxis(result, -1, axis)


def _get_nonlinear_constraint_violations(
    config: EnOptConfig, constraints: NDArray[np.float64]
) -> NDArray[np.float64]:
    assert config.nonlinear_constraints is not None
    return _get_constraint_violations(constraints, config.nonlinear_constraints.types)


def _get_constraint_violations(
    constraint_values: NDArray[np.float64], types: NDArray[np.ubyte]
) -> NDArray[np.float64]:
    le_inx = types == ConstraintType.LE
    ge_inx = types == ConstraintType.GE
    constraint_values = constraint_values.copy()
    constraint_values[le_inx] = np.maximum(constraint_values[le_inx], 0.0)
    constraint_values[ge_inx] = np.minimum(constraint_values[ge_inx], 0.0)
    return np.abs(constraint_values)


@overload
def _immutable_copy(data: NDArray[Any]) -> NDArray[Any]: ...


@overload
def _immutable_copy(data: Optional[NDArray[Any]]) -> Optional[NDArray[Any]]: ...


def _immutable_copy(data: Optional[NDArray[Any]]) -> Optional[NDArray[Any]]:
    if data is not None:
        data = data.copy()
        data.setflags(write=False)
    return data



import copy
from typing import Any, Callable, Dict, Optional

import numpy as np
import pytest

from ropt.config.enopt import (
    EnOptConfig,
    GradientConfig,
    LinearConstraintsConfig,
    NonlinearConstraintsConfig,
    ObjectiveFunctionsConfig,
    RealizationsConfig,
    VariablesConfig,
)
from ropt.enums import BoundaryType, ConstraintType, PerturbationType, VariableType


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "initial_values": np.array([1, 2]),
        },
        "objective_functions": {
            "weights": [1.0],
        },
        "optimizer": {
            "method": "dummy",
        },
    }


def test_check_variable_names() -> None:
    config = {"names": ["a", "b"], "initial_values": np.array([1, 2])}
    variables = VariablesConfig.model_validate(config)
    assert variables.names == ("a", "b")

    config = {"names": [1, "b"], "initial_values": np.array([1, 2])}
    variables = VariablesConfig.model_validate(config)
    assert variables.names is not None
    assert isinstance(variables.names[0], int)
    assert isinstance(variables.names[1], str)

    config = {"names": [("a",), ("b",)], "initial_values": np.array([1, 2])}
    variables = VariablesConfig.model_validate(config)
    assert variables.names == (("a",), ("b",))

    config = {"names": [(1, "a"), "b"], "initial_values": np.array([1, 2])}
    variables = VariablesConfig.model_validate(config)
    assert variables.names is not None
    assert isinstance(variables.names[0], tuple)
    assert isinstance(variables.names[0][0], int)
    assert isinstance(variables.names[0][1], str)
    assert isinstance(variables.names[1], str)

    config = {"names": ["a", "a", "b", "b", "c"], "initial_values": 0}
    with pytest.raises(ValueError, match="duplicate names: a, b"):
        VariablesConfig.model_validate(config)


def test_check_variable_arrays() -> None:
    config = {"names": ["a", "b"], "initial_values": np.array([1, 2])}

    for key in ["initial_values", "lower_bounds", "upper_bounds"]:
        config_copy = copy.deepcopy(config)

        variables = VariablesConfig.model_validate(config_copy)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(variables, key)[0] = 0

        config_copy[key] = np.array(0.0)
        variables = VariablesConfig.model_validate(config)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2

        config_copy[key] = np.array([0.0])
        variables = VariablesConfig.model_validate(config)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2

        config_copy[key] = np.array([0.0, 0.0, 0.0])
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            VariablesConfig.model_validate(config_copy)

        config_copy[key] = np.array([[0.0], [0.0]])
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            VariablesConfig.model_validate(config_copy)


def test_check_variable_arrays_no_names() -> None:
    config: Dict[str, Any] = {"initial_values": np.array([1, 2])}

    for key in ["initial_values", "lower_bounds", "upper_bounds"]:
        config_copy = copy.deepcopy(config)

        variables = VariablesConfig.model_validate(config_copy)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(variables, key)[0] = 0

        config_copy[key] = np.array(0.0)
        variables = VariablesConfig.model_validate(config)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2

        config_copy[key] = np.array([0.0])
        variables = VariablesConfig.model_validate(config)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2


def test_check_variable_convert_array() -> None:
    config = {"names": ["a", "b"], "initial_values": [1, 2]}

    for key in ["initial_values", "lower_bounds", "upper_bounds"]:
        config_copy = copy.deepcopy(config)

        variables = VariablesConfig.model_validate(config_copy)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2

        config_copy[key] = 0.0
        variables = VariablesConfig.model_validate(config)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2

        config_copy[key] = [0.0]
        variables = VariablesConfig.model_validate(config)
        assert getattr(variables, key).ndim == 1
        assert len(getattr(variables, key)) == 2

        config_copy[key] = [0.0, 0.0, 0.0]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            VariablesConfig.model_validate(config_copy)

        config_copy[key] = [[0.0], [0.0]]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            VariablesConfig.model_validate(config_copy)


def test_check_variable_arrays_types() -> None:
    config: Dict[str, Any] = {"initial_values": np.array([1, 2])}
    variables = VariablesConfig.model_validate(config)
    assert variables.types is None

    config["types"] = VariableType.INTEGER
    variables = VariablesConfig.model_validate(config)
    assert variables.types is not None
    assert np.all(variables.types == [VariableType.INTEGER, VariableType.INTEGER])

    config["types"] = [VariableType.INTEGER, VariableType.REAL]
    variables = VariablesConfig.model_validate(config)
    assert variables.types is not None
    assert np.all(variables.types == [VariableType.INTEGER, VariableType.REAL])


def test_get_formatted_names() -> None:
    config: Dict[str, Any] = {"initial_values": [0, 0]}
    variables = VariablesConfig.model_validate(config)
    joined = variables.get_formatted_names()
    assert joined is None

    config = {"names": ["a", ("b", "c"), "d"], "initial_values": 0}
    variables = VariablesConfig.model_validate(config)
    joined = variables.get_formatted_names()
    assert joined == ("a", "b:c", "d")

    config = {"names": [("a", "b", "c", "d")], "initial_values": 0}
    variables = VariablesConfig.model_validate(config)
    joined = variables.get_formatted_names()
    assert joined == ("a:b:c:d",)

    config = {"names": [("a", "b", "c", "d")], "initial_values": 0, "delimiters": "123"}
    variables = VariablesConfig.model_validate(config)
    joined = variables.get_formatted_names()
    assert joined == ("a1b2c3d",)

    config = {
        "names": [("a", "b", "c", "d")],
        "initial_values": 0,
        "delimiters": "1234",
    }
    variables = VariablesConfig.model_validate(config)
    joined = variables.get_formatted_names()
    assert joined == ("a1b2c3d",)

    config = {"names": [("a", "b", "c", "d")], "initial_values": 0, "delimiters": "12"}
    variables = VariablesConfig.model_validate(config)
    joined = variables.get_formatted_names()
    assert joined == ("a1b2c2d",)

    config = {"names": [("a", "b", "c", "d")], "initial_values": 0, "delimiters": ""}
    variables = VariablesConfig.model_validate(config)
    joined = variables.get_formatted_names()
    assert joined == ("abcd",)


def test_check_objective_function_arrays() -> None:
    config: Dict[str, Any] = {"names": ["a", "b"]}

    for key in ["scales", "weights"]:
        config_copy = copy.deepcopy(config)

        objective_functions = ObjectiveFunctionsConfig.model_validate(config_copy)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(objective_functions, key)[0] = 0

        config_copy[key] = np.array(0.0)
        objective_functions = ObjectiveFunctionsConfig.model_validate(config)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2

        config_copy[key] = np.array([0.0])
        objective_functions = ObjectiveFunctionsConfig.model_validate(config)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2

        config_copy[key] = np.array([0.0, 0.0, 0.0])
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            ObjectiveFunctionsConfig.model_validate(config_copy)

        config_copy[key] = np.array([[0.0], [0.0]])
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            ObjectiveFunctionsConfig.model_validate(config_copy)

    config_copy = copy.deepcopy(config)
    config_copy["weights"] = np.array([1.0, 1.0])
    objective_functions = ObjectiveFunctionsConfig.model_validate(config_copy)
    assert objective_functions.weights.sum() == 1.0


def test_check_objective_function_arrays_no_names() -> None:
    config: Dict[str, Any] = {"weights": [1, 1]}

    for key in ["scales", "weights"]:
        config_copy = copy.deepcopy(config)

        objective_functions = ObjectiveFunctionsConfig.model_validate(config_copy)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(objective_functions, key)[0] = 0

        config_copy[key] = np.array(0.0)
        objective_functions = ObjectiveFunctionsConfig.model_validate(config)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2

        config_copy[key] = np.array([0.0])
        objective_functions = ObjectiveFunctionsConfig.model_validate(config)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2

    config_copy = copy.deepcopy(config)
    config_copy["weights"] = np.array([1.0, 1.0])
    objective_functions = ObjectiveFunctionsConfig.model_validate(config_copy)
    assert objective_functions.weights.sum() == 1.0


def test_check_objective_function_convert_arrays() -> None:
    config: Dict[str, Any] = {"names": ["a", "b"]}

    for key in ["scales", "weights"]:
        config_copy = copy.deepcopy(config)

        objective_functions = ObjectiveFunctionsConfig.model_validate(config_copy)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2

        config_copy[key] = 0.0
        objective_functions = ObjectiveFunctionsConfig.model_validate(config)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2

        config_copy[key] = [0.0]
        objective_functions = ObjectiveFunctionsConfig.model_validate(config)
        assert getattr(objective_functions, key).ndim == 1
        assert len(getattr(objective_functions, key)) == 2

        config_copy[key] = [0.0, 0.0, 0.0]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            ObjectiveFunctionsConfig.model_validate(config_copy)

        config_copy[key] = [[0.0], [0.0]]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            ObjectiveFunctionsConfig.model_validate(config_copy)

    config_copy = copy.deepcopy(config)
    config_copy["weights"] = [1.0, 1.0]
    objective_functions = ObjectiveFunctionsConfig.model_validate(config_copy)
    assert objective_functions.weights.sum() == 1.0


def test_check_linear_constraints() -> None:
    config = {
        "coefficients": np.array([[1, 2], [1, 2], [1, 2]]),
        "rhs_values": np.array([1, 2, 3]),
        "types": [ConstraintType.LE, ConstraintType.GE, ConstraintType.EQ],
    }
    linear_constraints = LinearConstraintsConfig.model_validate(config)
    assert linear_constraints.coefficients is not None
    with pytest.raises(ValueError):  # noqa: PT011
        linear_constraints.coefficients[0, 0] = 0
    with pytest.raises(ValueError):  # noqa: PT011
        linear_constraints.rhs_values[0] = 0


def test_check_linear_constraints_convert() -> None:
    config = {
        "coefficients": [[1, 2], [1, 2], [1, 2]],
        "rhs_values": [1, 2, 3],
        "types": [ConstraintType.LE, ConstraintType.GE, ConstraintType.EQ],
    }
    LinearConstraintsConfig.model_validate(config)


def test_check_linear_constraints_vector_shapes() -> None:
    config = {
        "coefficients": [[1, 2, 3], [1, 2, 3]],
        "rhs_values": [1, 2],
        "types": [ConstraintType.LE, ConstraintType.GE],
    }
    LinearConstraintsConfig.model_validate(config)

    config_copy = copy.deepcopy(config)
    config_copy["rhs_values"] = [1, 2, 3]
    with pytest.raises(
        ValueError,
        match="rhs_values cannot be broadcasted to a length of 2",
    ):
        LinearConstraintsConfig.model_validate(config_copy)

    config_copy = copy.deepcopy(config)
    config_copy["types"] = [ConstraintType.LE, ConstraintType.GE, ConstraintType.EQ]
    with pytest.raises(
        ValueError,
        match="types cannot be broadcasted to a length of 2",
    ):
        LinearConstraintsConfig.model_validate(config_copy)


def test_check_nonlinear_constraint_arrays() -> None:
    config = {
        "names": ["a", "b"],
        "types": [ConstraintType.EQ, ConstraintType.LE],
        "rhs_values": np.array([1.0, 1.0]),
    }

    for key in ["rhs_values", "scales"]:
        config_copy = copy.deepcopy(config)

        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config_copy)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(nonlinear_constraints, key)[0] = 0

        config_copy[key] = np.array(1.0)
        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2

        config_copy[key] = np.array([1.0])
        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2

        config_copy[key] = np.array([1.0, 1.0, 1.0])
        with pytest.raises(
            ValueError,
            match=f"{key} cannot be broadcasted to a length of 2",
        ):
            NonlinearConstraintsConfig.model_validate(config_copy)

        config_copy[key] = np.array([[1.0], [1.0]])
        with pytest.raises(
            ValueError,
            match=f"{key} cannot be broadcasted to a length of 2",
        ):
            NonlinearConstraintsConfig.model_validate(config_copy)


def test_check_nonlinear_constraint_arrays_no_names() -> None:
    config = {
        "types": [ConstraintType.EQ, ConstraintType.LE],
        "rhs_values": np.array([1.0, 1.0]),
    }

    for key in ["rhs_values", "scales"]:
        config_copy = copy.deepcopy(config)

        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config_copy)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(nonlinear_constraints, key)[0] = 0

        config_copy[key] = np.array(1.0)
        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2

        config_copy[key] = np.array([1.0])
        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2


def test_check_nonlinear_constraint_convert_arrays() -> None:
    config = {
        "names": ["a", "b"],
        "types": [ConstraintType.EQ, ConstraintType.LE],
        "rhs_values": [1.0, 1.0],
    }

    for key in ["rhs_values", "scales"]:
        config_copy = copy.deepcopy(config)

        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config_copy)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2

        config_copy[key] = 1.0
        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2

        config_copy[key] = [1.0]
        nonlinear_constraints = NonlinearConstraintsConfig.model_validate(config)
        assert getattr(nonlinear_constraints, key).ndim == 1
        assert len(getattr(nonlinear_constraints, key)) == 2

        config_copy[key] = [1.0, 1.0, 1.0]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            NonlinearConstraintsConfig.model_validate(config_copy)

        config_copy[key] = [[1.0], [1.0]]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            NonlinearConstraintsConfig.model_validate(config_copy)


def test_check_realization_names() -> None:
    config = {"names": [1, "2"]}
    realizations = RealizationsConfig.model_validate(config)
    assert realizations.names is not None
    assert isinstance(realizations.names[0], int)
    assert isinstance(realizations.names[1], str)


def test_check_realization_arrays() -> None:
    config: Dict[str, Any] = {"names": ["a", "b"]}

    for key in ["weights"]:
        config_copy = copy.deepcopy(config)

        realizations = RealizationsConfig.model_validate(config_copy)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(realizations, key)[0] = 0

        config_copy[key] = np.array(0.0)
        realizations = RealizationsConfig.model_validate(config)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2

        config_copy[key] = np.array([0.0])
        realizations = RealizationsConfig.model_validate(config)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2

        config_copy[key] = np.array([0.0, 0.0, 0.0])
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            RealizationsConfig.model_validate(config_copy)

        config_copy[key] = np.array([[0.0], [0.0]])
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            RealizationsConfig.model_validate(config_copy)


def test_check_realization_arrays_no_names() -> None:
    config: Dict[str, Any] = {"weights": [1, 1]}

    for key in ["weights"]:
        config_copy = copy.deepcopy(config)

        realizations = RealizationsConfig.model_validate(config_copy)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2
        with pytest.raises(ValueError):  # noqa: PT011
            getattr(realizations, key)[0] = 0

        config_copy[key] = np.array(0.0)
        realizations = RealizationsConfig.model_validate(config)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2

        config_copy[key] = np.array([0.0])
        realizations = RealizationsConfig.model_validate(config)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2


def test_check_realization_convert_arrays() -> None:
    config: Dict[str, Any] = {"names": ["a", "b"]}

    for key in ["weights"]:
        config_copy = copy.deepcopy(config)

        realizations = RealizationsConfig.model_validate(config_copy)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2

        config_copy[key] = 0.0
        realizations = RealizationsConfig.model_validate(config)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2

        config_copy[key] = [0.0]
        realizations = RealizationsConfig.model_validate(config)
        assert getattr(realizations, key).ndim == 1
        assert len(getattr(realizations, key)) == 2

        config_copy[key] = [0.0, 0.0, 0.0]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            RealizationsConfig.model_validate(config_copy)

        config_copy[key] = [[0.0], [0.0]]
        with pytest.raises(
            ValueError, match=f"{key} cannot be broadcasted to a length of 2"
        ):
            RealizationsConfig.model_validate(config_copy)


def test_check_perturbations() -> None:
    GradientConfig()
    gradients = GradientConfig(perturbation_magnitudes=np.array([0.1]))
    assert gradients.perturbation_magnitudes == np.array([0.1])


def test_check_config(enopt_config: Any) -> None:
    EnOptConfig.model_validate(enopt_config)


def test_check_config_linear_constraints(enopt_config: Any) -> None:
    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 2, 3], [2, 3, 4]],
        "rhs_values": [1, 2],
        "types": [ConstraintType.GE, ConstraintType.GE],
    }
    with pytest.raises(
        ValueError,
        match="the coefficients matrix should have 2 columns",
    ):
        EnOptConfig.model_validate(enopt_config)


def test_check_config_perturbations(enopt_config: Any) -> None:
    enopt_config["gradient"] = {
        "perturbation_magnitudes": [1] * 2,
        "boundary_types": [BoundaryType.TRUNCATE_BOTH] * 2,
        "perturbation_types": [PerturbationType.ABSOLUTE] * 2,
    }
    EnOptConfig.model_validate(enopt_config)

    config_copy = copy.deepcopy(enopt_config)
    config_copy["gradient"]["perturbation_magnitudes"] = [1] * 3
    with pytest.raises(
        ValueError,
        match="the perturbation magnitudes cannot be broadcasted to a length of 2",
    ):
        EnOptConfig.model_validate(config_copy)

    config_copy = copy.deepcopy(enopt_config)
    config_copy["gradient"]["boundary_types"] = [BoundaryType.TRUNCATE_BOTH] * 3
    with pytest.raises(
        ValueError, match="perturbation boundary_types must have 2 items"
    ):
        EnOptConfig.model_validate(config_copy)

    config_copy = copy.deepcopy(enopt_config)
    config_copy["gradient"]["perturbation_types"] = [PerturbationType.ABSOLUTE] * 3
    with pytest.raises(ValueError, match="perturbation types must have 2 items"):
        EnOptConfig.model_validate(config_copy)


def test_check_config_min_success(enopt_config: Any) -> None:
    def gen_config(pert_min: Optional[int], real_min: Optional[int]) -> Dict[str, Any]:
        config: Dict[str, Any] = copy.deepcopy(enopt_config)
        config["realizations"] = {"weights": 4 * [1.0]}
        config["gradient"] = {}
        if pert_min is not None:
            config["gradient"]["perturbation_min_success"] = pert_min
        if real_min is not None:
            config["realizations"]["realization_min_success"] = real_min
        return config

    pert_test_map = {None: 5, 1: 1, 4: 4, 7: 5}
    real_test_map = {None: 4, 1: 1, 3: 3, 7: 4}
    test_space = zip(pert_test_map.keys(), real_test_map.keys())
    for pert_in, real_in in test_space:
        config = EnOptConfig.model_validate(gen_config(pert_in, real_in))
        assert pert_test_map[pert_in] == config.gradient.perturbation_min_success
        assert real_test_map[real_in] == config.realizations.realization_min_success


def test_perturbation_types(enopt_config: Any) -> None:
    enopt_config["gradient"] = {
        "perturbation_magnitudes": [0.1, 0.01],
        "perturbation_types": [PerturbationType.ABSOLUTE, PerturbationType.RELATIVE],
    }
    enopt_config["variables"]["lower_bounds"] = [0.0, np.inf]
    enopt_config["variables"]["upper_bounds"] = [1.0, 600.0]
    with pytest.raises(
        ValueError,
        match="The variable bounds must be finite to use relative perturbations",
    ):
        config = EnOptConfig.model_validate(enopt_config)

    enopt_config["variables"]["initial_values"] = [0.0, 0.0, 0.0]
    enopt_config["variables"]["lower_bounds"] = [0.0, 100.0, 0.0]
    enopt_config["variables"]["upper_bounds"] = [np.inf, 600.0, 1.0]
    enopt_config["variables"]["scales"] = [1.0, 1.0, 50]
    enopt_config["gradient"] = {
        "perturbation_magnitudes": [0.1, 0.01, 1.0],
        "perturbation_types": [
            PerturbationType.ABSOLUTE,
            PerturbationType.RELATIVE,
            PerturbationType.SCALED,
        ],
    }
    config = EnOptConfig.model_validate(enopt_config)
    assert np.allclose(config.gradient.perturbation_magnitudes, [0.1, 5.0, 0.02])


def test_config_inputs_property(
    enopt_config: Any, assert_equal_dicts: Callable[[Any, Any], None]
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    assert_equal_dicts(enopt_config, config.original_inputs)



"""The optimization event class."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional, Set, Tuple

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig
    from ropt.enums import EventType, OptimizerExitCode
    from ropt.results import Results


@dataclass
class Event:
    """The `Event` class stores optimization event data.

    While running an optimization plan, callbacks can be connected to react to
    events triggered during execution. These callbacks accept a single `Event`
    object containing information about the event.

    The actual data contained in the object depends on the nature of the event.
    Refer to the documentation of the [`EventType`][ropt.enums.EventType]
    enumeration for more details.

    Attributes:
        event_type: The type of the event.
        config:     The current configuration object of the executing plan.
        results:    Optional results passed with the event.
        exit_code:  An optional exit code.
        tags:       Optional tags added to the results.
        plan_id:    The ID of the plan that generated the event.
    """

    event_type: EventType
    config: EnOptConfig
    results: Optional[Tuple[Results, ...]] = None
    exit_code: Optional[OptimizerExitCode] = None
    tags: Set[str] = field(default_factory=set)
    plan_id: Tuple[int, ...] = field(default_factory=tuple)



from typing import Any, Dict

import numpy as np
import pytest

from ropt.enums import EventType, OptimizerExitCode
from ropt.plan import BasicOptimizer, Event
from ropt.results import FunctionResults

pytestmark = [pytest.mark.slow]


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "optimizer": {
            "method": "external/slsqp",
            "tolerance": 1e-4,
            "max_iterations": 50,
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def test_external_run(enopt_config: Any, evaluator: Any) -> None:
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0, 0, 0.5], atol=0.02)


def test_external_max_functions_exceeded(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 100

    def track_results(event: Event) -> None:
        nonlocal last_evaluation
        assert event.results
        assert isinstance(event.results[0].result_id, int)
        last_evaluation = event.results[0].result_id

    max_functions = 2
    enopt_config["optimizer"]["max_functions"] = max_functions
    optimizer = BasicOptimizer(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, track_results
    )
    optimizer.run()
    assert last_evaluation == max_functions
    assert optimizer.exit_code == OptimizerExitCode.MAX_FUNCTIONS_REACHED


def test_external_failed_realizations(enopt_config: Any, evaluator: Any) -> None:
    def _observer(event: Event) -> None:
        assert event.results
        assert isinstance(event.results[0], FunctionResults)
        assert event.results[0].functions is None

    functions = [lambda _0, _1: np.array(1.0), lambda _0, _1: np.array(np.nan)]
    optimizer = BasicOptimizer(enopt_config, evaluator(functions)).add_observer(
        EventType.FINISHED_EVALUATION, _observer
    )
    optimizer.run()
    assert optimizer.exit_code == OptimizerExitCode.TOO_FEW_REALIZATIONS


def test_external_user_abort(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 100

    def _observer(event: Event) -> None:
        nonlocal last_evaluation
        assert event.results
        assert isinstance(event.results[0].result_id, int)
        last_evaluation = event.results[0].result_id
        if event.results[0].result_id == 1:
            optimizer.abort_optimization()

    optimizer = BasicOptimizer(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _observer
    )
    optimizer.run()
    assert optimizer.results is not None
    assert last_evaluation == 1
    assert optimizer.exit_code == OptimizerExitCode.USER_ABORT


def test_external_error(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["optimizer"]["options"] = {"ftol": "foo"}
    with pytest.raises(
        RuntimeError,
        match="External optimizer error: could not convert string to float: 'foo'",
    ):
        BasicOptimizer(enopt_config, evaluator()).run()



"""Write optimization results to tabular files."""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Iterable, Literal, Optional

from ._data_frame import ResultsDataFrame
from ._utils import _HAVE_PANDAS, _HAVE_TABULATE, _extract_columns, _write_table

if TYPE_CHECKING:
    from pathlib import Path

    from ropt.config.enopt import EnOptConfig
    from ropt.results import Results

    if _HAVE_PANDAS:
        pass


class ResultsTable(ResultsDataFrame):
    """Generate files containing tables of optimization results.

    This class derives from the
    [`ResultsDataFrame`][ropt.report.ResultsDataFrame] class and writes the
    generated data frame in a tabular format to a text file.
    """

    def __init__(
        self,
        columns: Dict[str, str],
        path: Path,
        *,
        table_type: Literal["functions", "gradients"] = "functions",
        min_header_len: Optional[int] = None,
    ) -> None:
        """Initialize a results table.

        The `columns` parameter specifies which results are to be exported. The
        keys of the `columns` dictionary correspond to the `fields` parameter of
        the [`ResultsDataFrame`][ropt.report.ResultsDataFrame] parent class. The
        values are the corresponding titles of the columns of the table that is
        generated. As described in the documentation of the parent class, a
        single field may generate multiple columns, each with unique names
        (i.e., variable names). These are handled by adding the name to the
        column name below the main title. As a result, the header may consist of
        multiple lines, and the number of lines may vary according to requested
        fields. For a consistent result, the minimum number of header lines can
        be specified via the `min_header_len` argument. When needed, blank lines
        will be added to reach the specified minimum number of header lines.

        Tip: Reading the generated file.
            The resulting table can be read using a reader that can handle
            fixed-width columns, such as the read_fwf function of pandas. However,
            the header will need to skip a number of header lines. The
            min_header_len argument can be used to set the minimum number of lines
            in the header. If the generated header has fewer lines than
            min_header_len, empty lines will be added. For example:

            ```py
            # For a table generated with `min_header_len=3`:
            results = pd.read_fwf(
                "results.txt",
                header=list(range(3)),
                skip_blank_lines=False,
                skiprows=[3],
            )
            ```

        Args:
            columns:        Mapping of column names for the results table.
            path:           Optional location of the result file.
            table_type:     Type of the table.
            min_header_len: Minimal number of header lines.

        Raises:
            NotImplementedError: If the pandas or tabulate modules are not
                                 available
        """
        if not (_HAVE_TABULATE and _HAVE_PANDAS):
            msg = "ResultsTable requires the `tabulate` and `pandas` modules"
            raise NotImplementedError(msg)

        super().__init__(set(columns), table_type=table_type)
        self._columns = columns
        self._path = path
        self._min_header_len = min_header_len

    def add_results(self, config: EnOptConfig, results: Iterable[Results]) -> bool:
        """Add results to the table.

        This method can be called directly from any observers connected to
        events that produce results.

        Args:
            config:  The configuration of the optimizer generating the results.
            results: The results to add.

        Returns:
            True if a result was added, else False
        """
        if super().add_results(config, results):
            frame = self._frame.reset_index()
            table = _extract_columns(frame, mapping=self._columns)
            _write_table(table, self._path, self._min_header_len)
            return True
        return False



"""This module implements the default repeat step."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from pydantic import BaseModel, ConfigDict

from ropt.config.plan import PlanStepConfig  # noqa: TCH001
from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.plan import Plan


class DefaultRepeatStep(PlanStep):
    """The default repeat step.

    This step executes a defined list of steps in sequence for a specified
    number of repetitions. Optionally, it can store the current iteration number
    in a variable for tracking purposes. Each iteration preserves the original
    order of steps, ensuring sequential execution rather than parallel.

    The repeat step uses the [`DefaultRepeatStepWith`]
    [ropt.plugins.plan._repeat.DefaultRepeatStep.DefaultRepeatStepWith]
    configuration class to interpret the `with` field of the
    [`PlanStepConfig`][ropt.config.plan.PlanStepConfig] in a plan configuration.
    """

    class DefaultRepeatStepWith(BaseModel):
        """Parameters used by the default repeat step.

        The `steps` parameter defines a sequence of actions that are executed
        repeatedly, with the number of repetitions determined by `iterations`.
        If `var` is specified, it will be set to the current iteration count
        before each round of steps, allowing for iteration-specific actions or
        tracking.

        Attributes:
            iterations: The number of repetitions to perform.
            steps:      The list of steps to repeat in each iteration.
            var:        Optional variable to update with the current iteration count.
        """

        iterations: int
        steps: List[PlanStepConfig]
        var: Optional[str] = None

        model_config = ConfigDict(
            extra="forbid",
            validate_default=True,
            frozen=True,
        )

    def __init__(self, config: PlanStepConfig, plan: Plan) -> None:
        """Initialize a default optimizer step.

        Args:
            config: The configuration of the step.
            plan:   The plan that runs this step.
        """
        super().__init__(config, plan)

        with_ = self.DefaultRepeatStepWith.model_validate(config.with_)
        self._num = with_.iterations
        self._steps = self.plan.create_steps(with_.steps)
        self._var = with_.var

    def run(self) -> None:
        """Run the steps repeatedly."""
        for idx in range(self._num):
            if self._var is not None:
                self.plan[self._var] = idx
            self.plan.run_steps(self._steps)
            if self.plan.aborted:
                break



from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Tuple, cast

import numpy as np
import pytest
from numpy.typing import NDArray

from ropt.evaluator import ConcurrentEvaluator, ConcurrentTask, EvaluatorContext
from ropt.plan import BasicOptimizer


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "optimizer": {
            "tolerance": 1e-5,
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "realizations": {
            "weights": [1.0] * 3,
            "realization_min_success": 1,
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
            "perturbation_min_success": 2,
        },
    }


def _run_functions(
    functions: Tuple[Callable[..., Any], ...],
    variables: NDArray[np.float64],
    fail: bool,
) -> NDArray[np.float64]:
    result = np.zeros(len(functions), dtype=np.float64)
    for idx, function in enumerate(functions):
        if fail:
            raise RuntimeError
        result[idx] = function(variables, None)
    return result


@dataclass
class TaskTestEvaluator(ConcurrentTask):
    future: Any

    def get_objectives(self) -> Optional[NDArray[np.float64]]:
        return cast(NDArray[np.float64], self.future.result())


class ConcurrentTestEvaluator(ConcurrentEvaluator):
    def __init__(
        self, functions: Tuple[Callable[..., Any], ...], fail_index: int = -1
    ) -> None:
        super().__init__(enable_cache=True, polling=0.0)

        self._executor = ThreadPoolExecutor(max_workers=4)
        self._functions = functions
        self._fail_index = fail_index
        self._tasks: Dict[int, ConcurrentTask]

    def launch(
        self,
        batch_id: int,  # noqa: ARG002
        job_id: int,
        variables: NDArray[np.float64],
        context: EvaluatorContext,  # noqa: ARG002
    ) -> Optional[ConcurrentTask]:
        if job_id == 0:
            self._tasks = {}
        task = TaskTestEvaluator(
            future=self._executor.submit(
                _run_functions,
                self._functions,
                variables,
                self._fail_index == job_id,
            ),
        )
        self._tasks[job_id] = task
        return task

    def monitor(self) -> None:
        for idx, task in self._tasks.items():
            if task.future is not None and task.future.exception() is not None:
                print(f"error in evaluation {idx}")  # noqa: T201

    def disable_functions(self) -> None:
        self._functions = (lambda _0, _1: 0.0, lambda _0, _1: 0.0)


def test_concurrent(enopt_config: Any, test_functions: Any) -> None:
    evaluator = ConcurrentTestEvaluator(test_functions)
    variables = BasicOptimizer(enopt_config, evaluator).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


def test_concurrent_exception(
    enopt_config: Any, test_functions: Any, capsys: Any
) -> None:
    evaluator = ConcurrentTestEvaluator(test_functions, fail_index=2)
    variables = BasicOptimizer(enopt_config, evaluator).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)
    captured = capsys.readouterr()
    assert "error in evaluation 2" in captured.out


def test_concurrent_cache(enopt_config: Any, test_functions: Any) -> None:
    evaluator = ConcurrentTestEvaluator(test_functions)

    variables1 = BasicOptimizer(enopt_config, evaluator).run().variables
    assert variables1 is not None
    assert np.allclose(variables1, [0.0, 0.0, 0.5], atol=0.02)

    # Disable the functions, now the evaluator fully relies on its cache:
    evaluator.disable_functions()

    variables2 = BasicOptimizer(enopt_config, evaluator).run().variables
    assert variables2 is not None
    assert np.all(variables1 == variables2)



from __future__ import annotations

from dataclasses import dataclass
from importlib.util import find_spec
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Final, Optional, Type, TypeVar, Union

from ._gradient_evaluations import GradientEvaluations
from ._gradients import Gradients
from ._realizations import Realizations

if TYPE_CHECKING:
    from ._result_field import ResultField

from ._results import Results

_HAVE_XARRAY: Final = find_spec("xarray") is not None
_HAVE_NETCDF: Final = find_spec("netCDF4") is not None

if _HAVE_XARRAY:
    from ._xarray import _from_dataset
if _HAVE_XARRAY and _HAVE_NETCDF:
    from ._xarray import _from_netcdf

TypeResults = TypeVar("TypeResults", bound="Results")


@dataclass
class GradientResults(Results):
    """The `GradientResults` class stores gradient related results.

    This contains  the following additional information:

    1. The results of the function evaluations for perturbed variables.
    2. The parameters of the realizations, such as weights for objectives and
       constraints, and realization failures.
    3. The gradients of the calculated objectives and constraints.

    Attributes:
        evaluations:  Results of the function evaluations.
        realizations: The calculated parameters of the realizations.
        gradients:    The calculated gradients.
    """

    evaluations: GradientEvaluations
    realizations: Realizations
    gradients: Optional[Gradients]

    @classmethod
    def from_netcdf(cls: Type[TypeResults], filename: Union[str, Path]) -> TypeResults:
        """Read results from a netCDF4 file.

        Use of this method requires that the `xarray` and `netCDF4` modules are
        installed.

        The filename is assumed to have an ".nc" extension, which will be added
        if not present.

        Args:
            filename: The name or path of the file to read.

        Raises:
            NotImplementedError: If the xarray or the netCDF4 module is not installed.

        Returns:
            The loaded result.

        Warning:
            This function is only available if `xarray` and `netCDF4` are installed.
        """
        if not _HAVE_XARRAY or not _HAVE_NETCDF:
            msg = (
                "The xarray and netCDF4 modules must be installed "
                "to use Results.from_netcdf"
            )
            raise NotImplementedError(msg)

        _types: Dict[str, Type[ResultField]] = {
            "evaluations": GradientEvaluations,
            "realizations": Realizations,
            "gradients": Gradients,
        }

        results = _from_netcdf(Path(filename), cls)
        kwargs: Dict[str, Any] = {
            key: _types[key](**_from_dataset(value)) if key in _types else value
            for key, value in results.items()
        }
        kwargs.update({key: None for key in _types if key not in results})
        return cls(**kwargs)



"""This module implements the functions provided by the default plan plugin."""

from __future__ import annotations

from typing import Any, Dict, Hashable, Sequence


class ExpressionFunctions:
    """This class provides functions for the expression evaluator.

    The methods in this class are returned by the `data` property of the
    [`DefaultPlanPlugin`][ropt.plugins.plan.default.DefaultPlanPlugin] class and
    added to the expression evaluator used during execution of plans.
    """

    @staticmethod
    def mkdict(keys: Sequence[Hashable], values: Sequence[Any]) -> Dict[Hashable, Any]:
        """Create a dictionary from lists of keys and values.

        Args:
            keys:   The keys of the dictionary.
            values: The values of the dictionary.

        Returns:
            A dictionary constructed from keys and values.
        """
        return dict(zip(keys, values))



"""Utility functions for use by optimizer plugins.

This module provides utility functions to validate supported constraints, filter
linear constraints, and to retrieve the list of supported optimizers.
"""

from pathlib import Path
from typing import Dict, Optional, Set

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig, LinearConstraintsConfig
from ropt.enums import ConstraintType


def filter_linear_constraints(
    config: LinearConstraintsConfig, variable_indices: NDArray[np.intc]
) -> LinearConstraintsConfig:
    """Filter unnecessary constraints from a linear constraint configuration.

    In the case that the optimizer only optimizes a sub-set of the variables,
    linear constraints that are only formed from the unused variables are
    superfluous. This utility function removes those constraints from a  linear
    configuration constraint.

    Args:
        config:           The linear configuration constraint.
        variable_indices: The indices of the variables used by the optimizer.

    Returns:
        The filtered linear constraint configuration.
    """
    # Keep rows that only contain non-zero values for the active variables:
    mask = np.ones(config.coefficients.shape[-1], dtype=np.bool_)
    mask[variable_indices] = False
    keep_rows = np.all(config.coefficients[:, mask] == 0, axis=1)
    types = config.types[keep_rows]
    coefficients = config.coefficients[keep_rows, :]
    rhs_values = config.rhs_values[keep_rows]
    # Keep coefficients for the active variables:
    coefficients = coefficients[:, variable_indices]

    return LinearConstraintsConfig(
        coefficients=coefficients,
        types=types,
        rhs_values=rhs_values,
    )


_MESSAGES = {
    "bounds": "bound constraints",
    "linear:eq": "linear equality constraints",
    "linear:ineq": "linear inequality constraints",
    "nonlinear:eq": "non-linear equality constraints",
    "nonlinear:ineq": "non-linear inequality constraints",
}


def validate_supported_constraints(
    config: EnOptConfig,
    method: str,
    supported_constraints: Dict[str, Set[str]],
    required_constraints: Dict[str, Set[str]],
) -> None:
    """Check if the requested optimization features are supported or required.

    The keys of the supported_constraints and required_constraints dicts specify
    the type of the constraint as shown in the example below. The values are
    sets of method names that support or require the type of constraint
    specified by the key.

    For example:
    {
        "bounds": {"L-BFGS-B", "TNC", "SLSQP"},
        "linear:eq": {"SLSQP"},
        "linear:ineq": {"SLSQP"},
        "nonlinear:eq": {"SLSQP"},
        "nonlinear:ineq": {"SLSQP"},
    }

    Args:
        config:                The ensemble optimizer configuration object.
        method:                The method to check.
        supported_constraints: Specify the supported constraints.
        required_constraints:  Specify the required constraints.
    """
    _validate_bounds(config, method, supported_constraints, required_constraints)
    _validate_linear_constraints(
        config, method, supported_constraints, required_constraints
    )
    _validate_nonlinear_constraints(
        config, method, supported_constraints, required_constraints
    )


def _check_constraint(
    constraint_type: str,
    method: str,
    supported_constraints: Dict[str, Set[str]],
    required_constraints: Dict[str, Set[str]],
    *,
    have_constraint: bool,
) -> None:
    supported = {
        algo.lower() for algo in supported_constraints.get(constraint_type, set())
    }
    required = {
        algo.lower() for algo in required_constraints.get(constraint_type, set())
    }
    msg = _MESSAGES[constraint_type]
    if have_constraint and method.lower() not in supported:
        msg = f"optimizer {method} does not support {msg}"
        raise NotImplementedError(msg)
    if not have_constraint and method.lower() in required:
        msg = f"optimizer {method} requires {msg}"
        raise NotImplementedError(msg)


def _validate_bounds(
    config: EnOptConfig,
    method: str,
    supported_constraints: Dict[str, Set[str]],
    required_constraints: Dict[str, Set[str]],
) -> None:
    _check_constraint(
        "bounds",
        method,
        supported_constraints,
        required_constraints,
        have_constraint=bool(
            np.isfinite(config.variables.lower_bounds).any()
            or np.isfinite(config.variables.upper_bounds).any(),
        ),
    )


def _validate_linear_constraints(
    config: EnOptConfig,
    method: str,
    supported_constraints: Dict[str, Set[str]],
    required_constraints: Dict[str, Set[str]],
) -> None:
    linear_constraints = config.linear_constraints
    if linear_constraints is None:
        return

    if config.variables.indices is not None:
        linear_constraints = filter_linear_constraints(
            linear_constraints, config.variables.indices
        )

    _check_constraint(
        "linear:ineq",
        method,
        supported_constraints,
        required_constraints,
        have_constraint=bool(
            np.any(linear_constraints.types == ConstraintType.LE)
            or np.any(linear_constraints.types == ConstraintType.GE),
        ),
    )

    _check_constraint(
        "linear:eq",
        method,
        supported_constraints,
        required_constraints,
        have_constraint=bool(np.any(linear_constraints.types == ConstraintType.EQ)),
    )


def _validate_nonlinear_constraints(
    config: EnOptConfig,
    method: str,
    supported_constraints: Dict[str, Set[str]],
    required_constraints: Dict[str, Set[str]],
) -> None:
    nonlinear_constraints = config.nonlinear_constraints
    if nonlinear_constraints is None:
        return

    _check_constraint(
        "nonlinear:ineq",
        method,
        supported_constraints,
        required_constraints,
        have_constraint=bool(
            np.any(nonlinear_constraints.types == ConstraintType.LE)
            or np.any(nonlinear_constraints.types == ConstraintType.GE),
        ),
    )

    _check_constraint(
        "nonlinear:eq",
        method,
        supported_constraints,
        required_constraints,
        have_constraint=bool(
            np.any(nonlinear_constraints.types == ConstraintType.EQ),
        ),
    )


def create_output_path(
    base_name: str,
    base_dir: Optional[Path] = None,
    name: Optional[str] = None,
    suffix: Optional[str] = None,
) -> Path:
    """Create an output path name.

    If the path already exists, an index is appended to it.

    Args:
        base_name: Base name of the path.
        base_dir:  Optional directory to base the path in.
        name:      Optional optimization step name to include in the name.
        suffix:    Optional suffix for the resulting path.

    Returns:
        The constructed path
    """
    if base_dir is not None:
        base_dir.mkdir(parents=True, exist_ok=True)
    if name is not None:
        base_name += f"-{name}"
    output = base_dir / base_name if base_dir is not None else Path(base_name)
    if suffix is not None:
        output = output.with_suffix(suffix)
    while output.exists():
        fields = base_name.split("-")
        if fields[-1].strip().isdigit():
            index = int(fields[-1]) + 1
            base_name = "-".join(fields[:-1]) + f"-{index:03}"
        else:
            base_name = f"{base_name}-001"
        output = base_dir / base_name if base_dir is not None else Path(base_name)
        if suffix is not None:
            output = output.with_suffix(suffix)
    return output



from functools import partial
from typing import Any, Dict

import pytest

from ropt.config.enopt import EnOptConfig
from ropt.enums import EventType
from ropt.plan import BasicOptimizer, Event
from ropt.report import ResultsDataFrame

# Requires pandas:
pd = pytest.importorskip("pandas")


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "speculative": True,
            "max_functions": 3,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
            "upper_bounds": 1.0,
            "lower_bounds": -1.0,
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def _handle_results(event: Event, reporter: ResultsDataFrame) -> None:
    assert event.results is not None
    reporter.add_results(event.config, event.results)


def test_dataframe_results_no_results(enopt_config: Any, evaluator: Any) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(set())
    BasicOptimizer(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter),
    ).run()
    assert reporter.frame.empty


def test_dataframe_results_function_results(enopt_config: Any, evaluator: Any) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "result_id",
            "evaluations.variables",
        },
    )
    BasicOptimizer(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", idx) for idx in range(3)
    ]


def test_dataframe_results_function_results_formatted_names(
    enopt_config: Any, evaluator: Any
) -> None:
    enopt_config["variables"]["names"] = [("a", 1), ("a", 2), ("a", 3)]
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "result_id",
            "evaluations.variables",
        },
    )
    BasicOptimizer(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", f"a:{idx}") for idx in range(1, 4)
    ]


def test_dataframe_results_gradient_results(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["names"] = [("a", 1), ("a", 2), ("a", 3)]
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "result_id",
            "gradients.weighted_objective",
        },
        table_type="gradients",
    )
    BasicOptimizer(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("gradients.weighted_objective", f"a:{idx}") for idx in range(1, 4)
    ]


def test_dataframe_results_metadata(enopt_config: Any, evaluator: Any) -> None:
    reporter = ResultsDataFrame(
        {
            "result_id",
            "evaluations.variables",
            "metadata.foo.bar",
            "metadata.not.existing",
        },
    )

    def even_handler(event: Event) -> None:
        assert event.results is not None
        assert event.results is not None
        for item in event.results:
            item.metadata["foo"] = {"bar": 1}
        reporter.add_results(EnOptConfig.model_validate(enopt_config), event.results)

    BasicOptimizer(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, even_handler
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", idx) for idx in range(3)
    ] + ["metadata.foo.bar"]



"""This module implements the default set step."""

from __future__ import annotations

import copy
import re
from collections.abc import Mapping, MutableMapping, MutableSequence
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union

from pydantic import ConfigDict, RootModel

from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.config.plan import PlanStepConfig
    from ropt.plan import Plan


class DefaultSetStep(PlanStep):
    """The default set step.

    Set steps are used to modify the contents of plan variables. They specify
    one or more variables to set, either as a dictionary of variable-name/value
    pairs, or as a list of such dictionaries.

    Variables containing dictionaries can be modified by specifying an entry
    using the `[]` operator. Similarly, variables containing objects with
    attributes can be modified using the `.` operator. These can be mixed and
    nested to any depth to modify complex variables. For example, the expression
    `$var['foo'].bar[0]` is valid if `var` contains a dict-like value with a
    `foo` entry that has a `bar` attribute containing a list.

    The evaluator step uses the [`DefaultSetStepWith`]
    [ropt.plugins.plan._set.DefaultSetStep.DefaultSetStepWith]
    configuration class to parse the `with` field of the
    [`PlanStepConfig`][ropt.config.plan.PlanStepConfig] used to specify this step
    in a plan configuration.

    Note: Dictionary vs Lists
        Multiple variables may be set in a single step, either by using a single
        dictionary of variable/value pairs or by providing a list of
        dictionaries. Both approaches are generally equivalent. However, if a
        dictionary is loaded from a JSON or YAML file, the order of the keys is
        not guaranteed. Keep this in mind if the variables are interdependent
        and the assignment order matters.
    """

    class DefaultSetStepWith(
        RootModel[Union[Dict[str, Any], Tuple[Dict[str, Any], ...]]]
    ):
        """Configuration for a set step.

        A set step is used to change the value of one or more plan variables.
        The parameters provided to the step should either be a dictionary of
        variable-name/value pairs, or a list of such dictionaries. When the set
        step is run by the plan, the variables are set to the given values.
        Variables that are referenced may be dictionaries or objects with
        attributes. These can be modified by using the `[]` or `.` operators
        (possibly nested) with the specified variable name.

        Info: Using Expressions in the Value
            Optionally, the supplied value may be an expression, following the rules
            of the [`eval`][ropt.plan.Plan.eval] method of the plan object executing
            the steps. Refer to the method's documentation for more information
            on supported expressions.

        Note: Dictionaries vs. Lists
            The parameters provided to the set step may be dictionaries or lists
            of dictionaries. Multiple variables can be set in this manner in
            both cases. However, care should be taken if a dictionary is used
            where the order of the keys is not well defined, for instance, if it
            has been read from a JSON file. If the order in which variables are
            set is important, it is recommended to use a list.
        """

    root: Union[Dict[str, Any], Tuple[Dict[str, Any], ...]]

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
        frozen=True,
    )

    def __init__(self, config: PlanStepConfig, plan: Plan) -> None:
        """Initialize a set step.

        Args:
            config: The configuration of the step.
            plan:   The plan that runs this step.
        """
        super().__init__(config, plan)
        expr = self.DefaultSetStepWith.model_validate(config.with_).root
        if isinstance(expr, Mapping):
            expr = (expr,)

        self._plan = plan
        self._targets: List[Any] = []
        self._names: List[str] = []
        self._keys: List[List[str]] = []
        self._values: List[Any] = []

        for list_item in expr:
            if not isinstance(list_item, Mapping):
                msg = "`set` must be called with a var/value dict or a list of var/value dicts"
                raise TypeError(msg)
            for target, value in list_item.items():
                pattern = re.findall(
                    r"([^\[^\.]+)|(\[.*?\])|(\.\b\w+\b)", target.strip()
                )
                name, *keys = [x.strip() for group in pattern for x in group if x]
                if name not in self._plan:
                    msg = f"Unknown variable name: {name}"
                    raise AttributeError(msg)
                self._targets.append(target)
                self._names.append(name)
                self._keys.append(
                    [
                        key[1:] if key.startswith(".") else "$(" + key[1:-1] + ")"
                        for key in keys
                    ]
                )
                self._values.append(value)

    def run(self) -> None:
        """Run the set step."""
        for target_string, name, keys, value in zip(
            self._targets, self._names, self._keys, self._values
        ):
            evaluated_value = copy.deepcopy(self._plan.eval(value))
            if not keys:
                self._plan[name] = evaluated_value
            else:
                target = self._plan[name]
                try:
                    for key in keys[:-1]:
                        target = self._get_target(target, key)
                    self._set_target(target, keys[-1], evaluated_value)
                except (AttributeError, KeyError):
                    msg = f"Invalid attribute access: {target_string}"
                    raise AttributeError(msg) from None

    def _get_target(self, target: Any, key: str) -> Any:  # noqa: ANN401
        if key.startswith("$("):
            if not isinstance(target, (MutableMapping, MutableSequence)):
                raise KeyError
            return target[self._plan.eval(key)]
        return getattr(target, key)

    def _set_target(self, target: Any, key: str, value: Any) -> None:  # noqa: ANN401
        if key.startswith("$("):
            if not isinstance(target, (MutableMapping, MutableSequence)):
                raise KeyError
            target[self._plan.eval(key)] = value
        else:
            setattr(target, key, value)
